#####################################################################################
# LMB_THERMO_FIT
#
# Binary mixture fitting routine
# Fit voltage and other data of binary mixtures
#
#####################################################################################


import numpy as np
from scipy.misc import derivative
from scipy.optimize import leastsq

#Get constants from const
import const
F = const.F #Faraday constant (C/mol)
R = const.R #Gas constant
z = const.z #Valence number



#####################################################################################
#Datasets
#######################################
#Gasior et. al. (1994) Bi-Li System. Thermodynamic Properties and the Phase Diagram Calculations
x_A_Gasior = np.array([0.010,0.015,0.025,0.050,0.075,0.100,0.125,0.150,0.175,0.200,0.225,0.250,0.275,0.300,0.325,0.350,0.375,0.400,0.425,0.450,0.475,0.500,0.525,0.941,0.942,0.948,0.954,0.970])
A_Gasior = np.array([856,844,831,807,807,801,791,787,785,781,783,774,777,774,780,771,770,767,762,737,719,682,663,20,20,11,3,1])/1000 #V
B_Gasior = np.array([0.255,0.229,0.195,0.158,0.118,0.096,0.084,0.068,0.054,0.041,0.023,0.017,-0.001,-0.010,-0.031,-0.034,-0.046,-0.054,-0.061,-0.045,-0.036,-0.009,0.000,0.006,0.003,0.002,0.003,0.001])/1000 #V/K
T_Gasior=826 # Table fix temperature
T_min_Gasior=773 # minimum Temperature for Gasior measurements
T_max_Gasior=975 # maximum Temperature for Gasior measurements
int_T_Gasior=10 #number of temperature points to be created (for single fit)
x_min_Gasior=0 # minimum value to be fitted (liquid region only)
x_max_Gasior=18 # maximum value to be fitted (liquid region only)
x_A_Gasior=x_A_Gasior[x_min_Gasior:x_max_Gasior]
T_Gasior=np.asarray(len(x_A_Gasior)*[T_Gasior])
A_Gasior=A_Gasior[x_min_Gasior:x_max_Gasior]
B_Gasior=B_Gasior[x_min_Gasior:x_max_Gasior]


#######################################
# Demidov und Morachevsky (1973) Thermodynamic Properties of Liquid Li-Bi Alloys
x_A_Demidov = np.array([0.028,0.045,0.090,0.114,0.144,0.160,0.225,0.300,0.360])
E_Demidov = np.array([0.957,0.925,0.877,0.861,0.843,0.835,0.803,0.770,0.741]) #V
dEdT_Demidov = np.array([24,20,14,12,10,9,5,1,-2])*1e-5 #V/K
T_Demidov=800 # Table fix temperature
T_min_Demidov=510+273 # minimum Temperature for Gasior measurements
T_max_Demidov=560+273 # maximum Temperature for Gasior measurements
int_T_Demidov=10 #number of temperature points to be created (for single fit)
x_min_Demidov=0 # minimum value to be fitted (liquid region only)
x_max_Demidov=9 # maximum value to be fitted (liquid region only)
x_A_Demidov=x_A_Demidov[x_min_Demidov:x_max_Demidov]
T_Demidov=np.asarray(len(x_A_Demidov)*[T_Demidov])
A_Demidov=E_Demidov[x_min_Demidov:x_max_Demidov]-T_Demidov*dEdT_Demidov[x_min_Demidov:x_max_Demidov]
B_Demidov=dEdT_Demidov[x_min_Demidov:x_max_Demidov]
#E_Demidov_ABT=A_Demidov+T_Demidov*B_Demidov


#######################################
# Foster et al. - 1964 - Thermodynamics of Binary Alloys. I. The Lithium-Bismuth System
x_A_Foster = np.array([0.01,0.02,0.05,0.1,0.15,0.2,0.25,0.300,0.35]) #Dummy values
T_Foster=(775+1100)/2# Table fix temperature
T_min_Foster=775 # minimum Temperature for Foster measurements
T_max_Foster=1100 # maximum Temperature for Foster measurements
int_T_Foster=10 # number of temperature points to be created (for single fit)
x_min_Foster=0 # minimum value to be fitted (liquid region only)
x_max_Foster=9 # maximum value to be fitted (liquid region only)
x_A_Foster=x_A_Foster[x_min_Foster:x_max_Foster]
T_Foster=np.asarray(len(x_A_Foster)*[T_Foster])

#Functions
def E_Foster(x,T): #Equation is not right, gives wrong values
    mu_E = -((9397 + 18.16 * T - 0.0109 * T ** 2) + (7103 - 19.44 * T + 0.0068 * T ** 2) * (x))  # cal/mol
    E_Li = (mu_E * 4.185 + R * T * np.log(x)) / -F  # Check correct Calorie to Joule conversion
    return E_Li
E_Foster_result=E_Foster(x_A_Foster ,T_Foster)

def g_Li_Foster(x,T):
    g = -F * E_Foster(x,T)/1000
    return g
g_Li_Foster_result=g_Li_Foster(x_A_Foster ,T_Foster)

def dEdT_Foster(x,T):
    def E_Li_Foster_dgdt(T,x):
        E=E_Foster(x,T)
        return E
    dEdT=derivative(E_Li_Foster_dgdt, T,args=[x], dx=1e-6)
    return dEdT
dEdT_Li_Foster_result=dEdT_Foster(x_A_Demidov ,T_Demidov)

def s_Li_Foster(x,T):
    s = F * dEdT_Foster(x,T)
    return s
s_Li_Foster_result=s_Li_Foster(x_A_Foster,T_Foster)
s_Li_Foster_Demidov_result=s_Li_Foster(x_A_Demidov,T_Demidov)
s_Li_Foster_Gasior_result=s_Li_Foster(x_A_Gasior,T_Gasior)

#Derived E=A+BT function parameters
def A_Foster(x,T):
    A=E_Foster(x,T)-dEdT_Foster(x,T)*T
    return A
A_Foster_result=A_Foster(x_A_Foster,T_Foster)

def B_Foster(x,T):
    B=dEdT_Foster(x,T)
    return B
B_Foster_result=B_Foster(x_A_Foster,T_Foster)

#Prepare arrays for x, T, A, B, E
interval_x_Foster_array=np.ones(int_T_Foster) #empty (filled with 1) x array in lenght of T
interval_T_Foster_array=np.ones(x_max_Foster-x_min_Foster)#empty (filled with 1) T array in lenght of x
T_range_T_Foster=np.linspace(T_min_Foster,T_max_Foster,int_T_Foster) #1D temperature array
T_range_Foster=interval_T_Foster_array*T_range_T_Foster[:, np.newaxis] #2D temperature array for all x
x_range_Foster=interval_x_Foster_array[:, np.newaxis]*x_A_Foster #2D concentration array for all T
A_Foster_fit=A_Foster(x_range_Foster,T_range_Foster) #parameter A for all xT in 2D
B_Foster_fit=B_Foster(x_range_Foster,T_range_Foster) #parameter B for all xT in 2D
x_Foster_new = np.reshape(x_range_Foster, (1, -1))
x_Foster_new= x_Foster_new.flatten() #x for all xT in 1D
T_Foster_new = np.reshape(T_range_Foster, (1, -1))
T_Foster_new= T_Foster_new.flatten() #T for all xT in 1D
A_Foster_new = np.reshape(A_Foster_fit, (1, -1))
A_Foster_new= A_Foster_new.flatten() #A in 1D
B_Foster_new = np.reshape(B_Foster_fit, (1, -1))
B_Foster_new= B_Foster_new.flatten() #B in 1D
E_Foster_new=A_Foster_new+B_Foster_new*T_Foster_new #E in 1D


########################################
#Foster et al.  1963 ANL-6802 p.59, Foster et al. 1964 Fig. 2
#T_Foster=np.array([751.1560,770.6380,790.9960,811.3480,831.6940,852.0390,872.3780,893.5860,914.7880,935.9960,957.1910,979.2690,1000.4580,1022.5230,1043.7050,1065.7630,1086.9390,1108.9910,1130.1540,1152.1920,816.585,826.539,840.613,854.687,868.766,881.665,894.569,902.196,916.274,929.765,944.436,958.523,972.601,986.101,1000.193,1014.872,1024.843,1038.948,1053.632,1067.732,1084.769,1098.874,1113.557,1127.662,749.8260,770.1390,791.3480,812.5490,833.7580,854.9590,876.1610,897.3560,918.5520,939.7410,961.8120,983.8760,1005.0590,1027.1170,1049.1750,1071.2270,1093.2790,1115.3300,1136.4870,1158.5380,751.1310,772.3390,793.5350,814.7300,836.8080,857.9960,880.0680,902.1320,923.3150,945.3730,966.5420,986.8290,1007.9920,1030.0370,1051.1940,1072.3500,1093.5070,1114.6630,1134.9240,1155.1920,814.518,828.014,853.852,867.944,904.937,918.446,931.963,937.249,951.928,994.825,1008.343,1035.373,1048.899,1094.159,1107.689,1130.03,1147.676,750.7360,771.9310,793.9960,816.0610,838.1190,860.1770,882.2420,903.4110,925.4630,946.6190,968.6640,989.8210,1010.9770,1032.1270,1053.2710,1074.4140,1095.5580,1115.8120,1136.0670,1155.4390,817.809,830.731,844.831,847.768,861.281,872.444,887.128,901.82,915.925,930.029,944.13,957.651,971.761,984.103,991.16,1005.273,1019.387,1032.321,1046.435,1059.374,1072.321,1085.256,1098.194,1111.129,1127.009,819.886,822.231,836.336,850.441,864.545,878.067,892.176,906.281,920.39,934.499,948.613,962.726,976.252,989.787,1003.905,1017.431,1030.374,1044.492,1056.852,1070.97,1083.913,1096.865,1109.813,1122.765,1130.413])
#E_Foster=np.array([0.9137,0.9214,0.9284,0.9347,0.9403,0.9458,0.9507,0.9542,0.9569,0.9604,0.9624,0.9645,0.9658,0.9665,0.9671,0.9670,0.9670,0.9662,0.9647,0.9625,0.933111,0.937369,0.940472,0.943575,0.946083,0.949176,0.951674,0.952933,0.955441,0.957943,0.959861,0.961178,0.963686,0.964997,0.965719,0.966446,0.968321,0.967256,0.967387,0.966918,0.966474,0.965409,0.96554,0.964475,0.8651,0.8672,0.8706,0.8734,0.8768,0.8796,0.8823,0.8844,0.8864,0.8878,0.8891,0.8898,0.8904,0.8903,0.8903,0.8895,0.8887,0.8879,0.8858,0.8850,0.8151,0.8185,0.8206,0.8226,0.8247,0.8260,0.8273,0.8280,0.8286,0.8286,0.8278,0.8270,0.8255,0.8241,0.8219,0.8197,0.8175,0.8153,0.8117,0.8089,0.814581,0.816488,0.817315,0.818036,0.819558,0.819678,0.818608,0.818655,0.819382,0.816788,0.815717,0.814172,0.811911,0.807551,0.804694,0.801916,0.7985,0.7721,0.7741,0.7748,0.7754,0.7753,0.7753,0.7759,0.7751,0.7743,0.7722,0.7707,0.7685,0.7663,0.7634,0.7598,0.7563,0.7527,0.7484,0.7441,0.7398,0.766372,0.766488,0.766018,0.766045,0.76557,0.765074,0.765205,0.764145,0.76308,0.762015,0.761546,0.75988,0.758219,0.757138,0.75601,0.753754,0.751498,0.749827,0.747571,0.745305,0.741847,0.740176,0.737909,0.736238,0.733403,0.723513,0.724129,0.723064,0.721999,0.720934,0.719268,0.717608,0.716542,0.714882,0.713221,0.710965,0.708709,0.706448,0.702996,0.700144,0.697883,0.695021,0.692169,0.688706,0.685855,0.682993,0.67894,0.675482,0.671429,0.669711])
#x_A_Foster=np.array([0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.05,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.11,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.19,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.3,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4,0.4])


#######################################
# Predel et al. - 1979 - Kalorimetrische Untersuchung flüssiger Lithium-Thallium-, Lithium-Indium- und Lithium-Wismut-Legierungen
x_B_Predel=np.array([0.009,0.0231,0.0423,0.0632,0.0879,0.0274,0.0684,0.9661,0.9302,0.8890,0.8490,0.8046,0.7548,0.7073,0.6664,0.6291,0.5845,0.5448,0.5099,0.4776,0.4462,0.4218,0.4124])
H_Predel=np.array([-2.03,-4.91,-10.02,-15.36,-22.04,-7.21,-18.06,-2.27,-4.52,-7.10,-9.83,-12.94,-16.43,-19.86,-22.49,-25.24,-28.77,-31.93,-35.34,-38.30,-40.76,-43.41,-44.38])
x_A_Predel=1-x_B_Predel
T_Predel=985 # Table fix temperature
x_min_Predel=7 # minimum value to be fitted (liquid region only)
x_max_Predel=16 # maximum value to be fitted (liquid region only)
x_A_Predel=x_A_Predel[x_min_Predel:x_max_Predel]
T_Predel=np.asarray(len(x_A_Predel)*[T_Predel])


#####################################################################################
# Set up combined E=A+BT fitting data
# Uses one temperature array for all fits!
# To do: Temperature for each dataset
x_A=np.hstack((x_A_Gasior,x_A_Demidov)) # Combine commposition values
A=np.hstack((A_Gasior,A_Demidov))   # combine A values
B=np.hstack((B_Gasior,B_Demidov))   # combine B Values
T_min=773 # minimum temperature for voltage array
T_max=975 # maximum temperature for voltage array
int_T=20 # number of temperature points to be created (for combined fit)


#####################################################################################
#Create voltage array via E=A+BT for given x and temperature range and reduce to 1D array
def E_ABT_Array(x_A,A,B,T_min,T_max,int_T):
    interval_T=int_T # number of Temperature points for fitting
    interval_x=len(x_A) #number of x values
    x_range=x_A
    interval_x_array=np.ones(interval_T) #Dimension for x in T
    interval_T_array=np.ones(interval_x) #Dimension for T in x
    T_range_T=np.linspace(T_min,T_max,interval_T) #Single temperature array
    T_range=interval_T_array*T_range_T[:, np.newaxis] #Multiple temperature array for all x
    x_range=interval_x_array[:, np.newaxis]*x_range #Multiple concentration array for all T
    E_range=(A+B*T_range) #Create Voltage Values for all x and T
    E_range=np.ravel(E_range)#Flatten array for use in curve_fit
    xT_range_Tuple=([x_range,T_range]) # Create xT Tuple list
    xT_range= np.asarray(xT_range_Tuple) #Change Tuple to array
    x_new = np.reshape(xT_range[0],(1,-1))
    x_new = x_new.flatten()
    T_new = np.reshape(xT_range[1],(1,-1))
    T_new = T_new.flatten()
    E_new = E_range
    return x_new,T_new,E_new # x, T, E sorted in one-dimensional arrays
#x_range_leastsq, T_range_leastsq, E_range_leastsq =  E_ABT_Array(x_A_Gasior,A_Gasior,B_Gasior,T_min_Gasior,T_max_Gasior,int_T_Gasior)
#x_range_leastsq, T_range_leastsq, E_range_leastsq =  E_ABT_Array(x_A_Demidov,A_Demidov,B_Demidov,T_min_Demidov,T_max_Demidov,int_T_Demidov)
x_range_leastsq, T_range_leastsq, E_range_leastsq =  E_ABT_Array(x_A,A,B,T_min,T_max,int_T)


#####################################################################################
#Add single data points (1D arrays) to array
x_range_leastsq = np.hstack((x_range_leastsq,x_Foster_new))
T_range_leastsq = np.hstack((T_range_leastsq,T_Foster_new))
E_range_leastsq = np.hstack((E_range_leastsq,E_Foster_new))


#####################################################################################
#Define fit functions for Voltage model
#Fitting with leastsq
def fitFunc_E_leastsq(x,T,o,p,a,b,c,d,e,f): #Combined fit function for E=f(A+BT)x
   AA = o + p * np.log(x)
   #BB = a * np.log(T) + b * x + c * x ** 2 + d * x ** 3 + e * x ** 4 + f * x * np.log(x)  #
   BB = a * np.log(x) + b * x + c * x ** 2 + d * x ** 3 + e * x ** 4 + f * x * np.log(x)  #
   E_fit = AA+T*BB #mV
   #E_fit=np.ravel(E_fit) #macht expected type
   #global Numberofcalls
   #Numberofcalls=Numberofcalls+1
   return E_fit

def integral_fitFunc_E(use, x="",T="",o="",p="",a="",b="",c="",d="",e="",f=""):
    if use=="integrate":
        import sympy as sym
        a = sym.Symbol('a')
        b = sym.Symbol('b')
        c = sym.Symbol('c')
        d = sym.Symbol('d')
        e = sym.Symbol('e')
        f = sym.Symbol('f')
        o = sym.Symbol('o')
        p = sym.Symbol('p')
        x = sym.Symbol('x')
        T = sym.Symbol('T')
        #Place fit function insympy notation here
        res=sym.integrate(sym.log(-(o + p * sym.log(x) + T * (a * sym.log(x) + b * x + c * x ** 2 + d * x ** 3 + e * x ** 4 + f * x * sym.log(x))* 96845 / 1000))/(1-x)**2, x)
        print(res) #copy the result from the console
    if use=="calculate":
        #insert result from console here
        int=T*c*x**3/3 + T*d*x**4/4 + T*e*x**5/5 + x**2*(T*b/2 - T*f/4) + x*(-T*a + o - p) + (T*a*x + T*f*x**2/2 + p*x)*np.log(x)
        return int

# Error function E_fit-E_meas=minimal
def ErrFunc_E_leastsq(params_leastsq_guess,x,T,E):
    E_fit=np.ravel(fitFunc_E_leastsq(x,T,*params_leastsq_guess))
    dE=E_fit-E
    return dE

#start parameters
params_leastsq_guess=np.asarray((1, 1, 1, 1, 1, 1, 1, 1)) #no guess
#params_leastsq_guess=np.asarray((745.58307673823595, -23.219002400843184, -0.058963023992130006, 1.3891830642733394, -4.8959007660674043, 7.0128027160130895, -4.5383904918879239, 0.58155524834100825)) #Gasior only

#Fitting
fitParams_E_leastsq, cov_fitParams, info_ls, msg, success = leastsq(ErrFunc_E_leastsq, params_leastsq_guess, args=(x_range_leastsq,T_range_leastsq, E_range_leastsq),ftol=1e-6, xtol=1e-6, full_output=True)

##Parameter sets for integration etc.
E_fit_args=(fitParams_E_leastsq[0],fitParams_E_leastsq[1],fitParams_E_leastsq[2],fitParams_E_leastsq[3],fitParams_E_leastsq[4],fitParams_E_leastsq[5],fitParams_E_leastsq[6],fitParams_E_leastsq[7]) # all argunments as list


#####################################################################################
#Fitted Voltage (active metal A)
E_result_Gasior=fitFunc_E_leastsq(x_A_Gasior, T_Gasior, *E_fit_args)
E_result_Demidov=fitFunc_E_leastsq(x_A_Demidov, T_Demidov, *E_fit_args)

##Integral fit function
#integral_fitFunc_E("integrate")
#result_int=integral_fitFunc_E("calculate",x_A_Demidov, T_Demidov, *E_fit_args)

# Print Fit results for copying
print(" ")
print("Fitting function: o + p * np.log(x) + T * (a * np.log(x) + b * x + c * x ** 2 + d * x ** 3 + e * x ** 4 + f * x * np.log(x))")
print("with: o, p, a, b, c, d, e, f")
print(E_fit_args)
print(" ")
