#####################################################################################
# LMB_THERMO_PLOT_3D
#
# Plot results from LMB_THERMO_CALC as 3D surface plot
#
#####################################################################################

import numpy
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d import axes3d
from scipy.interpolate import UnivariateSpline

#Get constants from const
import const
F = const.F #Faraday constant (C/mol)
R = const.R #Gas constant
z = const.z #Valence number

from fit import fitFunc_E_leastsq
from fit import E_fit_args



#####################################################################################
#3D plot of voltage surface

#Equilibrium potential
def E_eq(x, T, *E_fit_args):
    E = fitFunc_E_leastsq(x, T, *E_fit_args)
    return E #V


## Prepare figure
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

ax.set_xlabel('Temperature (K)')
ax.set_xlim3d(1450,500)
ax.set_ylabel('Lithium fraction (-)')
ax.set_ylim3d(0,1)
ax.set_zlabel('Open Circuit Voltage Li-Bi Cell (V)')
ax.set_zlim3d(0,2)

## Liquidus Line of Li-Bi from Sangster und Pelton - 1991 - The Bi-Li (Bismuth-Lithium) System
Liquidus = numpy.array([[-0.1, 544.0],  # Dummy value for x=0
                [0.0, 544.0],  # Exact #1
                [2.5, 539], #Dummy for spline
                [5.8423915, 532.9571],
                [10.0, 524.3147408], #Dummy for spline
                [14.0, 516.0],  # Exact #3
                [20.788044, 577.65234],
                [26.222826, 622.34766],
                [31.52174, 657.5621],
                [35.326088, 673.0],
                [37.0, 688.0],  # Exact #8
                [41.711956, 757.7878],
                [47.010868, 832.2799],
                [52.038044, 908.1264],
                [58.016304, 1006.99774],
                [61.005436, 1059.8195],
                [66.304344, 1191.1964],
                [71.603264, 1340.1805],
                [75.0, 1418.0],  # Exact #16
                [79.61957, 1313.0925],
                [84.23913, 1199.3228],
                [87.63587, 1096.3883],
                [90.48913, 1001.58014],
                [92.934784, 904.06323],
                [94.70109, 798.41986],
                [96.059784, 694.1309],
                [97.28261, 576.298],
                [99.0, 448.0],  # Exact
                [100, 453.7],  # Exact
                [101, 453.7]])  # Dummy value for x=100
Liquidus=Liquidus*[0.01,1] ## percent to fraction
Liquidus[:, 0], Liquidus[:, 1] = Liquidus[:, 1], Liquidus[:, 0].copy() ## switch columns

## Generate spline functions for liquidus line x_Li=0...0.75
TliqSpline=UnivariateSpline(Liquidus[1:18, 1], Liquidus[1:18, 0]) ## call by: TliqSpline(x_Li)
xliqSpline1=UnivariateSpline(-Liquidus[1:5, 0], Liquidus[1:5, 1]) ## call by: xliqSpline(T) x_Li<x_Li(eutectic) # new spline function needs increasing x-values --> use negative Temp.
xliqSpline2=UnivariateSpline(Liquidus[5:18, 0], Liquidus[5:18, 1]) ## call by: xliqSpline(T) x_Li>x_Li(eutectic)


## Generate test values for plot of temperature and composition
#m=27 ## No. of temperature values coarse
#n=41 ## No. of composition values coarse
#m=96## No. of temperature values medium
#n=51 ## No. of composition values medium
m=191## No. of temperature values fine
n=101 ## No. of composition values fine
#m=476## No. of temperature values extrafine
#n=251 ## No. of composition values extrafine

Tfit= numpy.linspace(500, 1450, m) # Temperature array for diagrams
#xfit= numpy.linspace(0, 1, n) # fully linear
#xfit1= numpy.logspace(.001, 4, num=n, base=.1) # fully logarithmic
xfit1= numpy.logspace(5, 1, num=int(n/2+.5), base=.1) # 0.00001...0.1 logarithmic
xfit2= numpy.linspace(0.1, 1, int(n/2-.5)) # 0.1...1 linear
xfit=numpy.concatenate((xfit1,xfit2),axis=0) # Concentration array for diagrams


## Plot temperature lines
All=numpy.empty([m,n,3])
for mm in range(0, m):
    for nn in range(0, n):
        Txfit=[Tfit[mm],xfit[nn]]
        xTfit = [xfit[nn],Tfit[mm]]
        Tl = TliqSpline(Txfit[1]) ##Liquidus Temperature for given composition
        if Tfit[mm]<516: ## Solid region: no voltage
            U = 0
        #elif xfit[nn] <= 0.05: ## Voltage for full cell
        #   #U = 1.2
        #   #U = -0.0000004333 * Tfit[mm] * Tfit[mm] + 0.0007300000 * Tfit[mm] + 0.8933333333  ## Test of dE/dT
        #   nue = -(9397 + 18.16 * Txfit[0] - 0.0109 * Txfit[0] * Txfit[0] + (
        #   7103 - 19.44 * Txfit[0] + 0.0068 * Txfit[0] *Txfit[0]) * Txfit[1])  ## Foster et al. 1964 eq. (3)
        #   U = (nue * 4.184 + 8.314459 * Txfit[0] * numpy.log(Txfit[1])) / -96485  ## Foster et al. 1964 eq. (2)"""
        #    U = fitFunc(Txfit, fitParams[0], fitParams[1], fitParams[2], fitParams[3], fitParams[4], fitParams[5],
        #            fitParams[6], fitParams[7], fitParams[8], fitParams[9], fitParams[10], fitParams[11])
        elif xfit[nn] > 0.75: ## Voltage for empty cell
            U = 0
        elif xfit[nn] > 0.73: ## Voltage for nearly empty cell !!! additional function for LiBi-LiBi3 needed!!!
            U1 = E_eq(xfit[nn], Tfit[mm], *E_fit_args)
            U = U1 + ((xfit[nn] - 0.73) / (0.75 - 0.73)) * (0 - U1)
        else: ## Voltage for cell according to fit function
            if xfit[nn]<0.14: ##x_Li<x_Li(eutectic)
               xl = xliqSpline1(-Txfit[0])
            else: ##x_Li>x_Li(eutectic)
                xl = xliqSpline2(Txfit[0])
            if Txfit[0]<Tl: ## Voltage in two-phase-region
                Tx_2P=[Txfit[0], xl]
                U = E_eq(xl, Txfit[0], *E_fit_args)
                if Txfit[0]<688 and Txfit[1]>0.5: ##LiBi phase
                    dU= -0.000359576*Txfit[0] + 0.246034967 ## Voltage step LiBi-Li3Bi
                else:
                    dU=0.0
                U=U-dU ## Voltage with step

            else: ## Voltage in liquid region
                U = E_eq(xfit[nn], Tfit[mm], *E_fit_args)

        All[mm][nn][0] = Tfit[mm]
        All[mm][nn][1] = xfit[nn]
        All[mm][nn][2] = U

    ax.plot(All[mm,:,0], All[mm,:,1], All[mm,:,2])


## Plot composition lines
All=numpy.empty([m,n,3])
for nn in range(0, n):
    for mm in range(0, m):
        Txfit=[Tfit[mm],xfit[nn]]
        Tl = TliqSpline(Txfit[1]) ##Liquidus Temperature for given composition
        if Tfit[mm] < 516: ## Solid region: no voltage
            U = 0
        #elif xfit[nn] <= 0.05:  ## Voltage for full cell
        #    #U = 1.2
        #    #U = -0.00000044 * Tfit[mm]  * Tfit[mm]  + 0.0007300000 * Tfit[mm]  + 0.8933333333  ## Test of dE/dT
        #    nue = -(9397 + 18.16 * Txfit[0] - 0.0109 * Txfit[0] * Txfit[0] + (
        #        7103 - 19.44 * Txfit[0] + 0.0068 * Txfit[0] * Txfit[0]) * Txfit[1])  ## Foster et al. 1964 eq. (3)
        #    U = (nue * 4.184 + 8.314459 * Txfit[0] * numpy.log(Txfit[1])) / -96485  ## Foster et al. 1964 eq. (2)"""
        #    U = fitFunc(Txfit, fitParams[0], fitParams[1], fitParams[2], fitParams[3], fitParams[4], fitParams[5],
        #            fitParams[6], fitParams[7], fitParams[8], fitParams[9], fitParams[10], fitParams[11])
        elif xfit[nn] > 0.75: ## Voltage for empty cell
            U=0
        elif xfit[nn] > 0.73:  ## Voltage for nearly empty cell !!! additional function for LiBi-LiBi3 needed!!!
            U1 = E_eq(xfit[nn], Tfit[mm], *E_fit_args)
            U = U1+((xfit[nn]-0.73)/(0.75-0.73))*(0-U1)
        else:  ## Voltage for cell according to fit function
            if xfit[nn] < 0.14:  ##x_Li<x_Li(eutectic)
                xl = xliqSpline1(-Txfit[0])
            else:  ##x_Li>x_Li(eutectic)
                xl = xliqSpline2(Txfit[0])
            if Txfit[0] < Tl:  ## Voltage in two-phase-region
                Tx_2P = [Txfit[0], xl]
                U = E_eq(xl, Txfit[0], *E_fit_args)
                if Txfit[0]<688 and Txfit[1]>0.5: ##LiBi phase
                    dU= -0.000359576*Txfit[0] + 0.246034967 ## Voltage step LiBi-Li3Bi
                else:
                    dU=0.0
                U=U-dU ## Voltage with step

            else:  ## Voltage in liquid region
                U = E_eq(xfit[nn], Txfit[0], *E_fit_args)

        All[mm][nn][0] = Tfit[mm]
        All[mm][nn][1] = xfit[nn]
        All[mm][nn][2] = U

    ax.plot(All[:,nn,0], All[:,nn,1], All[:,nn,2])

## Rotate Angle of View
ax.view_init(azim=70)

## Liquidus line in diagram
Uliq=[] ##Initialize voltage
for ll in range(1, 19):
    Ul=E_eq(Liquidus[ll, 1], Liquidus[ll, 0], *E_fit_args)
    Uliq.append(Ul)
Uliq[0]=1.2 ##Add Voltage for pure elements
Uliq=numpy.asarray(Uliq)
ax.plot(Liquidus[1:19,0], Liquidus[1:19,1],Uliq[0:18],'k-') ##Plot liquidus line


#Plot raw data used for fitting
from fit import x_range_leastsq
from fit import T_range_leastsq
from fit import E_range_leastsq

ax.plot(T_range_leastsq, x_range_leastsq, E_range_leastsq, color='k',linewidth=0, marker='+',markersize=5) ##Plot raw data

from fit import x_range_leastsq

fig.savefig("f.eps")






## Prepare second figure
fig = plt.figure()
ax2 = fig.add_subplot(111, projection='3d')
ax2.set_xlim3d(1450,500)
ax2.set_ylim3d(0,1)
ax2.set_zlim3d(0,2)
ax2.view_init(azim=70)

## Full labelling with numbers
#ax2.set_xlabel('Temperature (K)')
#ax2.set_ylabel('Lithium fraction (-)')
#ax2.set_zlabel('Open Circuit Voltage Li-Bi Cell (V)')

## Labeling with/without numbers
 ## Add labels
ax2.set_xlabel('Temperature')
ax2.set_ylabel('Lithium Fraction')
ax2.set_zlabel('Open Circuit Voltage Li-Bi Cell')
#ax2.grid(False)
#ax2.set_axis_off() ## Only surface

 ## Hide panes
ax2.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
ax2.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
ax2.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

 ## Hide spines
ax2.w_xaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
ax2.w_yaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
ax2.w_zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))

 ## Ticks and labels
ax2.set_xticks([500,750,1000,1250])
ax2.set_yticks([0.25,0.5,0.75,1])
ax2.set_zticks([0,0.4,0.6,0.8,1,1.2,1.4,1.6,1.8,2])
ax2.tick_params(color='w') ## hide tick lines
#ax2.tick_params(labelsize=18)
ax2.xaxis.labelpad=30
ax2.yaxis.labelpad=30
ax2.zaxis.labelpad=30


## Prepare surface plot
X = numpy.hstack(All[:,:,0])
Y =numpy.hstack(All[:,:,1])
Z =numpy.hstack(All[:,:,2])
ax2.plot_trisurf(X, Y, Z, vmin=0.5, vmax=1.2, cmap=cm.nipy_spectral, linewidth=0.0)
ax2.plot(Liquidus[1:19,0], Liquidus[1:19,1],Uliq[0:18],'0.75') ##Plot liquidus line


## Save figure
#plt.savefig('Name.eps', format='eps', dpi=1000)


plt.show()
