
import sys, numpy as np, scipy.constants, matplotlib.pyplot as plt
sys.path.append("../lib")
from emf import emf_x
from LiBi import *
from density import cToX
import density



################################################################################
# define materials 
################################################################################

# index '0' refers to start of charge, '1' to start of discharge

# define dimensions
d1 = 0.029		# diameter bottom
d2 = 0.013
A1 = np.pi*d1*d1/4.0	# surface area
hE1 = 0.0035            # electrolyte layer height, start of discharge

# define operation parameters
I = 1.                  # current
j = I/A1                # current density
T = 460			# °C
t0 = 1800.12
t1 = 3600.76
t2 = t0+t1

# define initial condition
#Idt = 3240.0            # initial discharge state (I * dt), not used
#n0 = Idt / (F*n_el)	# initial moles of Li, not used
nBi = 0.1		# initial amount of Bi

# define discretisation
nz = 100                                        # cells in z
deltaT = 0.1                                    # time step
tEnd = 30*60
tPoints = int(tEnd/deltaT + 1)

# define material
D = 4.0e-9              # diffusion coefficient (approximate)
n_el = 1                # number of electrons
F = scipy.constants.value('Faraday constant')

# read data
d = np.genfromtxt('0024_cycl_0_1A_0_-1A_30min_30min_30min_30min_1x.txt', skip_header=1)

################################################################################
# get the initial values for charge
################################################################################

# get x from emf, calc density, volume and c0

print("\n*************** Initial values charge ***************\n")
print("T in °C: ",T,'\n')
# the initial charge is not correct (Paolo)
# therefore, we get the initial x by looking on the voltage
#x0 = 0.251  # Paolo's value
x0 = 0.2728 # my value
print("x0 (estimated): ", x0, '\n')

print("emf0, measured", d[0][1]/1000.)
print("emf0, calculated", emf_x(np.array([x0]), T)[0])
print(" ")

n0 = x0*nBi / (1-x0)
print("n0 (Li): ", n0)
print("n0 (Bi): ", nBi, '\n')

import density 
rho = density.rhoLiBi_x(x0, T)
print("rho:  ", rho, '\n')

# volume is needed for concentration

V0 = (n0*MLi() + nBi*MBi())/rho
h0 = V0 / A1
print("V: ",  V0)
print("h: ", h0, '\n')
print("h(Bi only): ", nBi*MBi()/(A1*density.rhoLiBi_x(0.0, T)))

# get concentrations
cBi = nBi / V0
print("c0 (Bi)", cBi)
c0 = n0 / V0
print("c0 (Li)", c0)

# compute x0 by formula
c0calc = rho * x0 / (x0*MLi() + (1.-x0)*MBi())
print("c0, calculated from x0: ", c0calc, '\n')
print(" ")



################################################################################
# get the initial values, start of discharge
################################################################################

print("\n*************** Initial values discharge ***************\n")

# get concentration change due to charge
Idt = I * tEnd		# 1A, 30min
n1 = Idt / (F*n_el)	# amount of substance Li
x1 = (n0-n1)/(n0-n1+nBi)
c1 = density.rhoLiBi_x(x1,T) * x1/ (x1*MLi() + (1.-x1)*MBi())

print("x (computed): ", x1)
print("c1 (computed): ", c1)
#c1g = 8400.
#print("c1 (guessed): ",c1g,'\n')
print(" ")
print("emf(measured): 0.83")
print("emf(x1, computed): ", emf_x(np.array([x1]), T)[0])
print(" ")

V1 = (n1*MLi() + nBi*MBi())/density.rhoLiBi_x(x1, T)    # get volume for layer thickness
h1 = V1 / A1
print("V: ",  V1)
print("h: ", h1, '\n')
print("h(electrolyte): ", hE1 - (h0-h1), '\n')



################################################################################
# plot measurement
################################################################################

helmholtzBlau          = (       0.,  88./255., 156./255.)
rossendorfOrange       = (207./255., 104./255., 0.)
helmholtzDunkelBlau    = (       0.,  62./255., 110./255.)
grau1                  = ( 81./255.,  81./255.,  81./255. )
grau2                  = (156./255., 156./255., 156./255.)
grau3                  = (185./255., 185./255., 185./255.)
gesundheit             = (212./255.,  45./255.,  18./255.)
strukturDerMaterie     = (207./255., 104./255.,   0.)
energie                = (230./255., 175./255.,  17./255.)
erdeUndUmwelt          = ( 20./255.,  77./255.,  40./255.)
schluesseltechnologien = (169./255., 181./255.,   9./255.)
verkehrUndWeltraum     = (       0., 162./255., 224./255.)

f, ax = plt.subplots()
f.set_size_inches(6,4)

ax.xaxis.label.set_size(12)
ax.yaxis.label.set_size(12)
ax.tick_params(axis='x', labelsize=10)
ax.tick_params(axis='y', labelsize=10)

plt.plot(d[:,0]/60., d[:,1], '-', color=grau3, linewidth=5, label='experiment') # measurement starts after 1 min



################################################################################
# get ohmic overpotential
################################################################################

from LiBi import sigmaLiClFI
from ohm import etaCylinder
sigma = sigmaLiClFI(T)
eta0 = etaCylinder(I, sigma, d1, d2, hE1 - (h0-h1)) 	# at start of charge cathode is a little thicker, and electrolyte therefore thinner
eta1 = etaCylinder(I, sigma, d1, d2, hE1) 		# electrolyte layer thickness is given at start of discharge
print("Conductivity: ", sigma)
print("Ohmic loss (start charge): ", eta0)
print("Ohmic loss (start discharge): ", eta1)



################################################################################
# plot charge curve
################################################################################

# start up -> start charge
t = np.array([0, t0, t0])
e0 = emf_x(cToX(c0,T), T)
e=np.array([e0, e0, e0+eta0])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')

# charge
t = np.linspace(0, tEnd, 100)
Idt = I * t             # charge
n = n0 - Idt / (F*n_el) # amount of Li
x = n / (n+nBi)         # molar fraction Li
c = density.rhoLiBi_x(x, T) * x/ (x*MLi() + (1.-x)*MBi()) # concentration Li

V = (n*MLi() + nBi*MBi())/density.rhoLiBi_x(x, T)    	# get volume for layer thickness
h = V / A1						# thickness cathode
hE = hE1 - (h - h1)					# Li is removed; cathode gets thinner, electrolyte thicker

e = emf_x(cToX(c, T), T) + etaCylinder(I, sigma, d1, d2, hE)
plt.plot((t0+t)/60., 1e3*e, linewidth=2, color='k')

# end of charge
t = np.array([t1, t1, t1+tEnd])
e1 = emf_x(cToX(c1, T), T)
e = np.array([e1+eta1, e1, e1])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')

# start of discharge
t = np.array([t1+tEnd, t1+tEnd])
e = np.array([e1, e1-eta1])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')




################################################################################
# plot analytical solutions
################################################################################

# plot analytical solution for infinite layer
from diffAnalytical import diffSemiInfinite
t = np.linspace(0.0, tEnd, 10)                # time
cA = c1 + diffSemiInfinite(D, h1, j, n_el, t, h1)
emf = emf_x(cToX(cA, T), T) - eta1
plt.plot((t+t2)/60.,1e3*emf, 'o', linewidth=1, label='analytical model (semi-infinite)', color=schluesseltechnologien, markeredgecolor='k')
#plt.plot((t+t2)/60.,1e3*emf, 'o', linewidth=1, fillstyle='none', color='k')


# plot analytical solution for finite layer
from diffAnalytical import diffFiniteLayer
N = 1000 # number of summations for integral
t = np.linspace(0.0, tEnd, 100)                # time
cB = c1 + diffFiniteLayer(D, h1, j, N, n_el, t, h1)
emf2 = emf_x(cToX(cB, T), T) - eta1
#t = np.append(t, [tEnd, 60*60]) # add two points when cell is switched off
#emf2 = np.append(emf2, [emf_x(cToX(cB[-1], T), T), emf_x(cToX(cB[-1], T), T)])
plt.plot((t+t2)/60., 1e3*emf2, '-', linewidth=2, label='analytical model (finite)', color=rossendorfOrange)



################################################################################
# plot numerical solutions
################################################################################

# plot numerical solution
from fvm import c_numerical
t = np.linspace(0.0, tEnd, tPoints)
#cC,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T)
#emfC = emf_x(cToX(cC[:,-1], T), T) - eta
#plt.plot((t+t2)/60., 1e3*emfC, linewidth=0.5, label='numerical')

# with volume change
t = np.linspace(0.0, tEnd, tPoints)
#cD,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True)
#deltaH = z[:,-1] - h1   # height change for each time step
#emfD = emf_x(cToX(cD[:,-1], T), T) - etaCylinder(I, sigma, d1, d2, hE1-deltaH)
#plt.plot((t+t2)/60., 1e3*emfD, '--', linewidth=0.5, label='vol. change', color=rossendorfOrange)

# with variable diffusivity
t = np.linspace(0.0, tEnd, tPoints)
#cE,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, varD=True)
#emfE = emf_x(cToX(cE[:,-1], T), T) - eta
#plt.plot((t+t2)/60., 1e3*emfE, linewidth=0.5, label='variable D', color=rossendorfOrange)

# with vol change and variable D
t = np.linspace(0.0, tEnd, tPoints)
cF,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True, varD=True)[:2]
deltaH = z[:,-1] - h1     # height change for each time step
emfF = emf_x(cToX(cF[:,-1], T), T) - etaCylinder(I, sigma, d1, d2, hE1-deltaH)
# get profile, when current is switched off
#ta = np.linspace(tEnd, 60*60, tPoints) # 30 to 60 min, starting at t2 (=discharge)
#cFa,z = c_numerical(D, z[-1,-1], 0.0, MLi(), n_el, tEnd, deltaT, nz, cF[-1,:], T, volChange=True, varD=True) # j=0; layer height=z; initial concentration cF
#t = np.append(t, ta)
#emfFa = emf_x(cToX(cFa[:,-1], T), T)
#emfF = np.append(emfF, emfFa)
plt.plot((t+t2)/60., 1e3*emfF, '-', linewidth=2, label='numerical model', color=helmholtzBlau)
print('h(electrolyte) at the end:', hE1-deltaH[-1], '\n')


################################################################################
# save
################################################################################

plt.xlim(20,130)

ax.set(xlabel=r"$t$ in min", ylabel=r"$E$ in mV")
plt.legend()

f.subplots_adjust(left=0.13,bottom=0.13,right=0.98,top=0.98,wspace=0.15,hspace=0.25)
f.savefig("f.pdf")
