import sys, numpy as np, scipy.constants, matplotlib.pyplot as plt
sys.path.append("../lib")
from emf import emf_x
from LiBi import *
from density import cToX
import density



################################################################################
# define materials and dimensions
################################################################################

# define dimensions
d1 = 0.090              # diameter bottom (Infos by Steffen)
d2 = 0.030              # upper diameter
A = np.pi*d1*d1/4.0     # surface area (bottom)
hE1 = 0.0028            # electrolyte layer height, start of discharge

# define operation parameters
I = 5.                  # current
j = I/A                 # current density
T = 420                 # °C
t0=60.9			# start charge
t1=1860			# start discharge

# define initial condition
nBi = 1.248             # initial amount of Bi (Info Steffen)

# define discretisation
nz = 100                                        # cells in z
deltaT = 0.1                                    # time step
tEnd = 30*60
tPoints = int(tEnd/deltaT + 1)

# define material
D = 7.0e-9              # diffusion coefficient
n_el = 1                # number of electrons
F = scipy.constants.value('Faraday constant')

# read data
d = np.genfromtxt('0201_cycl_5A_-5A_30min_30min.txt', skip_header=1)



################################################################################
# get the initial values, start of charge
################################################################################

# get x from emf, calc density, volume and c0

print("\n*************** Initial values charge ***************\n")
print("T in °C: ", T, '\n')
x0 = 0.0945
print("x0 (estimated): ", x0, '\n')

print("emf0, measured", d[0][1]/1000.)
print("emf0, calculated", emf_x(np.array([x0]), T)[0])
print(" ")

n0 = x0*nBi / (1.-x0)
print("n0 (Li): ", n0)
print("n0 (Bi): ", nBi,'\n')

rho = density.rhoLiBi_x(x0,T)
print("rho:  ", rho, '\n')

# volume is needed for concentration

V0 = (n0*MLi() + nBi*MBi())/rho
h0 = V0 / A
print("V: ",  V0)
print("h: ", h0, '\n')
print("h (Bi only):", nBi*MBi()/(density.rhoLiBi_x(0.0, T)*A))

# get concentrations
cBi = nBi / V0
print("c0 (Bi)", cBi)
c0 = n0 / V0
print("c0 (Li)", c0)

# compute x0 by formula
c0calc = rho * x0 / (x0*MLi() + (1.-x0)*MBi())
print("c0, calculated from x0: ", c0calc, '\n')
print(" ")



################################################################################
# get the initial values, start of discharge
################################################################################

print("\n*************** Initial values discharge ***************\n")

# get concentration change due to charge
Idt = I * tEnd       	# 30 min
n1 = Idt / (F*n_el)     # amount of Li entering the cathode during discharge
x1 = (n0-n1)/(n0-n1+nBi)# molar fraction Li
c1 = density.rhoLiBi_x(x1,T) * x1/ (x1*MLi() + (1.-x1)*MBi()) # concentration Li

print("x (computed): ",x1)
print("c (computed): ",c1)
print(" ")
print("emf(x1, computed): ", emf_x(np.array([x1]), T)[0])
print(" ")

V1 = (n1*MLi() + nBi*MBi())/density.rhoLiBi_x(x1, T)    # get volume for layer thickness
h1 = V1 / A
print("V: ",  V1)
print("h: ", h1, '\n')
hE0 = hE1 - (h0-h1)					# height electrolyte at very beginning
print("h(electrolyte): ", hE0)



################################################################################
# plot measurement
################################################################################

helmholtzBlau          = (       0.,  88./255., 156./255.)
rossendorfOrange       = (207./255., 104./255., 0.)
helmholtzDunkelBlau    = (       0.,  62./255., 110./255.)
grau1                  = ( 81./255.,  81./255.,  81./255. )
grau2                  = (156./255., 156./255., 156./255.)
grau3                  = (185./255., 185./255., 185./255.)
gesundheit             = (212./255.,  45./255.,  18./255.)
strukturDerMaterie     = (207./255., 104./255.,   0.)
energie                = (230./255., 175./255.,  17./255.)
erdeUndUmwelt          = ( 20./255.,  77./255.,  40./255.)
schluesseltechnologien = (169./255., 181./255.,   9./255.)
verkehrUndWeltraum     = (       0., 162./255., 224./255.)

f, ax = plt.subplots()
f.set_size_inches(6,4)

ax.xaxis.label.set_size(12)
ax.yaxis.label.set_size(12)
ax.tick_params(axis='x', labelsize=10)
ax.tick_params(axis='y', labelsize=10)


plt.plot(d[:,0]/60., d[:,1], color=grau3, linewidth=5, label='experiment')



################################################################################
# get ohmic overpotential
################################################################################

from LiBi import sigmaLiClFI
from ohm import etaCylinder
sigma = sigmaLiClFI(T)
eta0 = etaCylinder(I, sigma, d1, d2, hE0)
eta1 = etaCylinder(I, sigma, d1, d2, hE1)
print("Conductivity: ", sigma)
print("Ohmic loss (start charge): ", eta0)
print("Ohmic loss (start discharge): ", eta1)



################################################################################
# plot charge curve
################################################################################

# start up -> start charge
t = np.array([0, t0, t0])
e0 = emf_x(cToX(c0, T), T)
e=np.array([e0, e0, e0+eta0])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')

# charge
t = np.linspace(0, tEnd, 100)
Idt = I * t       	# charge
n = n0 - Idt / (F*n_el) # amount of Li
x = n / (n+nBi)         # molar fraction Li
c = density.rhoLiBi_x(x, T) * x/ (x*MLi() + (1.-x)*MBi()) # concentration Li

V = (n*MLi() + nBi*MBi())/density.rhoLiBi_x(x, T)       # get volume for layer thickness
h = V / A                                               # thickness cathode
hE = hE1 - (h - h1)                                     # Li is removed; cathode gets thinner, electrolyte thicker


e = emf_x(cToX(c, T), T) + etaCylinder(I, sigma, d1, d2, hE)
plt.plot((t0+t)/60., 1e3*e, linewidth=2, color='k')

# end of charge
t = np.array([t1, t1])
e1 = emf_x(cToX(c1, T), T)
e = np.array([e1+eta1, e1])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')

# start of discharge
t = np.array([t1, t1])
e = np.array([e1, e1-eta1])
plt.plot(t/60., 1e3*e, linewidth=2, color='k')



################################################################################
# plot analytical solutions
################################################################################

# plot analytical solution for infinite layer
from diffAnalytical import diffSemiInfinite
t = np.linspace(0.0, tEnd, 6)                # time
cA = c1 + diffSemiInfinite(D, h1, j, n_el, t, h1)
emf = emf_x(cToX(cA,T),T) - eta1
plt.plot((t+t1)/60.,1e3*emf,'o',linewidth=1,label='analytical model (semi-infinite)', color=schluesseltechnologien, markeredgecolor='k')
#plt.plot((t+t1)/60.,1e3*emf,'o',linewidth=1,fillstyle='none', color='k')


# plot analytical solution for finite layer
from diffAnalytical import diffFiniteLayer
N = 1000 # number of summations for integral
t = np.linspace(0.0, tEnd, 100)                # time
cB = c1 + diffFiniteLayer(D, h1, j, N, n_el, t, h1)
emfB = emf_x(cToX(cB, T), T) - eta1
plt.plot((t+t1)/60., 1e3*emfB, '-', linewidth=2, label='analytical model (finite)', color=rossendorfOrange)

# charge
t = np.linspace(0, tEnd, 100)
Idt = I * t             # charge
n = n0 - Idt / (F*n_el) # amount of Li
x = n / (n+nBi)         # molar fraction Li
c = density.rhoLiBi_x(x, T) * x/ (x*MLi() + (1.-x)*MBi()) # concentration Li
e = emf_x(cToX(c, T), T) + eta1
plt.plot(np.array([t1+30*60, t1+30*60])/60., 1e3*np.array([emfB[-1], e[0]]), linewidth=2, color=rossendorfOrange)
plt.plot((t0+(60*60)+t)/60., 1e3*e, linewidth=2, color=schluesseltechnologien)



################################################################################
# plot numerical solutions
################################################################################

# plot numerical solution
from fvm import c_numerical
t = np.linspace(0.0, tEnd, tPoints)
#cC,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T)
#emfC = emf_x(cToX(cC[:,-1],T),T) - eta*hE1
#plt.plot((t+t1)/60.,1e3*emfC,linewidth=0.5,label='numerical')

# with volume change
t = np.linspace(0.0, tEnd, tPoints)
#cD,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True)
#deltaH = z[:,-1] - h1   # height change for each time step
#emfD = emf_x(cToX(cD[:,-1],T),T) - eta*(hE1-deltaH)
#plt.plot((t+t1)/60.,1e3*emfD,'--',linewidth=1,label='vol. change')

# with variable diffusivity
t = np.linspace(0.0, tEnd, tPoints)
#cE,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, varD=True)
#emfE = emf_x(cToX(cE[:,-1],T),T) - eta*hE1
#plt.plot((t+t1)/60.,1e3*emfE,linewidth=0.5,label='variable D')

# with vol change and variable D
t = np.linspace(0.0, tEnd, tPoints)
cF,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True, varD=True)[:2]
deltaH = z[:,-1] - h1     # height change for each time step
emfF = emf_x(cToX(cF[:,-1], T), T) - etaCylinder(I, sigma, d1, d2, hE1-deltaH)
plt.plot((t+t1)/60., 1e3*emfF, '-', linewidth=2, label='numerical solution', color=helmholtzBlau)
print('h(electrolyte) at the end:', hE1-deltaH[-1], '\n')

# charge
t = np.linspace(0, tEnd, 100)
Idt = I * t             # charge
n = n0 - Idt / (F*n_el) # amount of Li
x = n / (n+nBi)         # molar fraction Li
c = density.rhoLiBi_x(x, T) * x/ (x*MLi() + (1.-x)*MBi()) # concentration Li

V = (n*MLi() + nBi*MBi())/density.rhoLiBi_x(x, T)       # get volume for layer thickness
h = V / A                                               # thickness cathode
hE = hE1 - (h - h1)                                     # Li is removed; cathode gets thinner, electrolyte thicker

e = emf_x(cToX(c, T), T) + etaCylinder(I, sigma, d1, d2, hE)
plt.plot(np.array([t1+30*60, t1+30*60])/60., 1e3*np.array([emfF[-1], e[0]]), linewidth=2, color=helmholtzBlau)
plt.plot((t0+(60*60)+t)/60., 1e3*e, linewidth=2, color=helmholtzBlau)



################################################################################
# save
################################################################################

#plt.ylim(770,900)
plt.xlim(0,75)

ax.set(xlabel=r"$t$ in min", ylabel=r"$E$ in mV")
plt.legend(loc='upper right')

f.subplots_adjust(left=0.12,bottom=0.13,right=0.99,top=0.98,wspace=0.15,hspace=0.25)
f.savefig("f.pdf")
