

import sys, numpy as np, scipy.constants, matplotlib.pyplot as plt
sys.path.append("../lib")
from emf import emf_x
from LiBi import *
from density import cToX
import density



################################################################################
# define materials and dimensions
################################################################################

# define dimensions
d1 = 0.15		# diameter bottom (Ning paper)
d2 = 0.1                # upper diameter (estimated)
A = np.pi*d1*d1/4.0     # surface area (bottom)
hE1 = 0.01              # electrolyte layer height, start of discharge

# define operation parameters
j = 0.3*1e4             # current density (fig. 6, Ning et al.)
I = j*A
T = 550                 # °C

# define material
F = scipy.constants.value('Faraday constant')
n_el = 1.               # number of electrons
D = 8.0e-9              # diffusion coefficient

# define initial condition
mBi = 455.28*1e-3	
nBi = mBi / MBi()
C = 3.*nBi*F            # capacity in As (Li3Bi is discharge state; 3xnBi needs to be transferred)

# define discretisation
nz = 10			# cells in z
deltaT = 130            # time step
tEnd = C/I
tPoints = int(tEnd/deltaT + 1)




################################################################################
# get the initial values, start of charge
################################################################################

# get x from emf, calc density, volume and c0

print("\n*************** Initial values discharge ***************\n")
print("T in °C: ", T)
print("I in A: ", I)
print("j in A/m2: ", j)
print("A in m2: ", A)
print("Theoretical capacity: ", C/60./60., 'Ah')
print("Discharge time: ", tEnd, ' s')
print(" ")

x0 = 0.0
print("x0 (estimated): ", x0, '\n')
print(" ")

n0 = x0*nBi / (1.-x0)
print("n0 (Li): ", n0)
print("n0 (Bi): ", nBi,'\n')

rho = density.rhoLiBi_x(x0, T)
print("rho: ", rho, '\n')

# volume is needed for concentration

V0 = (n0*MLi() + nBi*MBi())/rho
h0 = V0 / A
print("V: ",  V0)
print("h: ", h0, '\n')
print("h (Bi only):", nBi*MBi()/(density.rhoLiBi_x(0.0, T)*A))
Vend = (3*nBi*MLi() + nBi*MBi())/density.rhoLiBi_x(0.75, T)
hend = Vend / A
print("h (end): ", hend)

# get concentrations
cBi = nBi / V0
print("c0 (Bi)", cBi)
c0 = n0 / V0
print("c0 (Li)", c0)

# compute x0 by formula
c0calc = rho * x0 / (x0*MLi() + (1.-x0)*MBi())
print("c0, calculated from x0: ", c0calc, '\n')
print(" ")



################################################################################
# plot measurement
################################################################################

helmholtzBlau          = (       0.,  88./255., 156./255.)
rossendorfOrange       = (207./255., 104./255., 0.)
helmholtzDunkelBlau    = (       0.,  62./255., 110./255.)
grau1                  = ( 81./255.,  81./255.,  81./255. )
grau2                  = (156./255., 156./255., 156./255.)
grau3                  = (185./255., 185./255., 185./255.)
gesundheit             = (212./255.,  45./255.,  18./255.)
strukturDerMaterie     = (207./255., 104./255.,   0.)
energie                = (230./255., 175./255.,  17./255.)
erdeUndUmwelt          = ( 20./255.,  77./255.,  40./255.)
schluesseltechnologien = (169./255., 181./255.,   9./255.)
verkehrUndWeltraum     = (       0., 162./255., 224./255.)

f, ax = plt.subplots()
f.set_size_inches(6,4)

ax.xaxis.label.set_size(12)
ax.yaxis.label.set_size(12)
ax.tick_params(axis='x', labelsize=10)
ax.tick_params(axis='y', labelsize=10)

plt.plot([C/3600.,C/3600.], [0,1e3], '--', color='k')



################################################################################
# get ohmic overpotential
################################################################################

from LiBi import sigmaLiClF75
from ohm import etaCylinder
sigma = sigmaLiClF75(T)
eta = etaCylinder(I, sigma, d1, d2, hE1)
print("Conductivity: ", sigma)
print("Ohmic loss: ", eta)



################################################################################
# plot emf
################################################################################

Q = np.linspace(0, 3.5*nBi*F, 1000, endpoint=True)      # charge, which can be transferred: 3*nBi*F
nLi = Q/F
x = nLi/(nLi+nBi)
plt.plot(Q/60./60., 1e3*emf_x(x, T), linewidth=2., color=grau2, label='open circuit potential')



################################################################################
# plot analytical solutions
################################################################################

# plot analytical solution for infinite layer
from diffAnalytical import diffSemiInfinite
t = np.linspace(0.0, tEnd, tPoints)                # time
cA = c0 + diffSemiInfinite(D, h0, j, n_el, t, h0)
emf = emf_x(cToX(cA, T), T) - eta
plt.plot(I*t/60./60., 1e3*emf, ':', linewidth=2, label='analytical (semi-infinite layer)', color=schluesseltechnologien)

# plot analytical solution for finite layer
from diffAnalytical import diffFiniteLayer
N = 1000 # number of summations for integral
t = np.linspace(0.0, tEnd, tPoints)                # time
cB = c0 + diffFiniteLayer(D, h0, j, N, n_el, t, h0)
emfB = emf_x(cToX(cB, T), T) - eta
plt.plot(I*t/60./60., 1e3*emfB, '-', linewidth=2, label='analytical (finite layer)', color=rossendorfOrange)



################################################################################
# plot numerical solutions
################################################################################

# plot numerical solution
import fdm, fvm
#t = np.linspace(0.0, tEnd, tPoints)
#cC,z = c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T)
#emfC = emf_x(cToX(cC[:,-1], T), T) - eta
#plt.plot(t/60., 1e3*emfC, linewidth=0.5, label='numerical')

# with volume change
t = np.linspace(0.0, tEnd, tPoints)
#cD,z = c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T, volChange=True)
#deltaH = z[:,-1] - h0   # height change for each time step
#emfD = emf_x(cToX(cD[:,-1],T),T) - etaCylinder(I,sigma,d1,d2,hE1-deltaH)
#plt.plot(t/60., 1e3*emfD, '--', linewidth=2, label='numerical (vol. change)', color=helmholtzBlau)

# with variable diffusivity
t = np.linspace(0.0, tEnd, tPoints)
#cE,z = c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T, varD=True)
#emfE = emf_x(cToX(cE[:,-1], T), T) - eta
#plt.plot(t/60., 1e3*emfE, ':', linewidth=2, label='numerical (var. D)', color=helmholtzBlau)

# with vol change and variable D
t = np.linspace(0.0, tEnd, tPoints)
cF,z,fc,dzF = fvm.c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T, volChange=True, varD=True)
deltaH = z[:,-1] - h0     # height change for each time step
emfF = emf_x(cToX(cF[:,-1], T), T) - etaCylinder(I,sigma,d1,d2,hE1-deltaH)
plt.plot(I*t/60./60., 1e3*emfF , linewidth=2, label='numerical (volume change)', color=helmholtzBlau)
print('h(electrolyte) at the end:', hE1-deltaH[-1], '\n')



################################################################################
# save
################################################################################

plt.ylim(0,1000)
plt.xlim(0,180)

plt.text(0.97*C/60./60.,800, r'Li$_3$Bi', rotation='vertical')

ax.set(xlabel=r"charge in Ah", ylabel=r"$E$ in mV")
plt.legend(loc='upper center')

f.subplots_adjust(left=0.12,bottom=0.11,right=0.98,top=0.98,wspace=0.15,hspace=0.25)
f.savefig("f.pdf")
