# figure 1
import sys, numpy as np, scipy.constants, matplotlib.pyplot as plt
sys.path.append("../lib")
from emf import emf_x
from LiBi import *
from density import cToX, xToC
import density



################################################################################
# define materials and dimensions
################################################################################

# define dimensions
d1 = 0.089              # diameter bottom 
d2 = 0.064              # upper diameter
A = np.pi*d1*d1/4.0     # surface area (bottom)
hE1 = 0.01              # electrolyte layer height, start of discharge

# define operation parameters
j = 0.3*1e4             # current density (fig. 6, Ning et al.)
I = j*A
T = 550                 # °C

# define initial condition
mBi = 166.8*1e-3	
nBi = mBi / MBi()

# define discretisation
nz = 20                                         # cells in z
deltaT = 130                                    # time step
tEnd = 65.*60*60/I
#tEnd = 3
tPoints = int(tEnd/deltaT + 1)
print(tPoints)

# define material
D = 8.0e-9              # diffusion coefficient
n_el = 1.               # number of electrons
F = scipy.constants.value('Faraday constant')

C = 3.*nBi*F            # capacity in As (Li3Bi is discharge state; 3xnBi needs to be transferred)



################################################################################
# get the initial values, start of charge
################################################################################

# get x from emf, calc density, volume and c0

print("\n*************** Initial values discharge ***************\n")
print("T in °C: ", T)
print("I in A: ", I)
print("j in A/m2: ", j)
print("A in m2: ", A)
print("Theoretical capacity: ", C/60./60., 'Ah')
print(" ")

x0 = 0.0
print("x0 (estimated): ", x0, '\n')
print('emf, measured: 1.05')	# see fig. 7a, Ning et al. (2015)
print("emf0, calculated", emf_x(np.array([x0]), T)[0])
print(" ")

n0 = x0*nBi / (1.-x0)
print("n0 (Li): ", n0)
print("n0 (Bi): ", nBi,'\n')

rho = density.rhoLiBi_x(x0, T)
print("rho: ", rho, '\n')

# volume is needed for concentration

V0 = (n0*MLi() + nBi*MBi())/rho
h0 = V0 / A
print("V: ",  V0)
print("h: ", h0, '\n')
print("h (Bi only):", nBi*MBi()/(density.rhoLiBi_x(0.0, T)*A))
Vend = (3*nBi*MLi() + nBi*MBi())/density.rhoLiBi_x(0.75, T)
hend = Vend / A
print("h (end): ", hend)

# get concentrations
cBi = nBi / V0
print("c0 (Bi)", cBi)
c0 = n0 / V0
print("c0 (Li)", c0)

# compute x0 by formula
c0calc = rho * x0 / (x0*MLi() + (1.-x0)*MBi())
print("c0, calculated from x0: ", c0calc, '\n')
print(" ")


################################################################################
# get ohmic overpotential
################################################################################

from LiBi import sigmaLiClF75
from ohm import etaCylinder#, hIntermetallic
sigma = sigmaLiClF75(T)
eta = etaCylinder(I, sigma, d1, d2, hE1)
print("Conductivity: ", sigma)
print("Ohmic loss: ", eta)



################################################################################
# plot measurement
################################################################################

f, ax = plt.subplots(3,4)
ax = ax.flatten()
f.set_size_inches(18,12)

color = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'k', 'k', 'k', 'k', 'k', 'k', 'dimgray', 'b', 'c']

#data = np.genfromtxt('data',skip_header=0)
#ax[0].plot(data[:,0], data[:,1], linewidth=2, color=color[10], label='experiment')
ax[0].plot([C/3600.,C/3600.], [0,1], '-', color='k', label='theoretical capacity')



################################################################################
# plot analytical solutions
################################################################################

# plot analytical solution for infinite layer
from diffAnalytical import diffSemiInfinite
#t = np.linspace(0.0, tEnd, tPoints)                # time
#cA = c0 + diffSemiInfinite(D, h0, j, n_el, t, h0)
#emf = emf_x(cToX(cA, T), T) - eta
#ax[0].plot(I*t/60./60., 1e3*emf, '-', linewidth=2, label='analytical (infinite layer)', color=grau2)

# plot analytical solution for finite layer
from diffAnalytical import diffFiniteLayer
N = 1000 # number of summations for integral
t = np.linspace(0.0, tEnd, tPoints)                # time
cB = c0 + diffFiniteLayer(D, h0, j, N, n_el, t, h0)
emfB = emf_x(cToX(cB, T), T) - eta
ax[0].plot(I*t/60./60., emfB, '-', linewidth=2, color=color[1], label='analytical (finite layer)')

 

################################################################################
# plot numerical solutions
################################################################################

# plot numerical solution
import fvm, fdm

# with volume change
t = np.linspace(0.0, tEnd, tPoints)

cD,z,fAll,dzAll = fvm.c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T, volChange=True)
deltaH = z[:,-1] - h0   # height change for each time step
emfD = emf_x(cToX(cD[:,-1], T), T) - etaCylinder(I,sigma,d1,d2,hE1-deltaH)

cDd,zd = fdm.c_numerical(D, h0, j, MLi(), n_el, tEnd, deltaT, nz, c0, T, volChange=True)
deltaHd = zd[:,-1] - h0   # height change for each time step
emfDd = emf_x(cToX(cDd[:,-1], T), T) - etaCylinder(I,sigma,d1,d2,hE1-deltaHd)

ax[0].plot(I*t/60./60., emfD, '-', linewidth=2, color=color[2], label='fvm')
ax[0].plot(I*t/60./60., emfDd, '-', linewidth=2, color=color[3], label='fdm')



#rhoM = density.cLiBi_rho(density.rhoLiBi_c(50000, T), T, 0)
#print(rhoM)



################################################################################
# plot emf
################################################################################

Q = np.linspace(0, 3.5*nBi*F, 1000, endpoint=True)	# charge, which can be transferred: 3*nBi*F
nLi = Q/F
x = nLi/(nLi+nBi)
ax[0].plot(Q/60./60., emf_x(x, T), color=color[12], label='emf')



################################################################################
# plot height of cathode layer
################################################################################

ax[1].plot([0,C/60./60], [1e3*h0, 1e3*hend], color=color[11], label='theory')
ax[1].plot(I*t/60./60., 1e3*z[:,-1], color=color[2], label='fvm')
ax[1].plot(I*t/60./60., 1e3*zd[:,-1], color=color[3], label='fdm')



################################################################################
# plot mole fraction x over time
################################################################################

ax[2].plot(I*t/60./60., cToX(cD[:,-1], T), color=color[2], label='fvm')
ax[2].plot(I*t/60./60., cToX(cDd[:,-1], T), color=color[3], label='fdm')
ax[2].plot(I*t/60./60., cToX(cB, T), color=color[1], label='analytical')
ax[2].plot(Q/60./60., x, color=color[12], label='emf')



################################################################################
# plot mole fraction x at last time step
################################################################################

ax[3].plot(cToX(cD[-1,:], T), 1e3*z[-1,:], color=color[2], label='fvm')
ax[3].plot(cToX(cDd[-1,:], T), 1e3*zd[-1,:], color=color[3], label='fdm')



################################################################################
# plot interface concentration over time
################################################################################

ax[4].plot(Q/60./60., 1e-4*xToC(x, T), color=color[12], label='emf')
ax[4].plot(I*t/60./60., 1e-4*cD[:,-1], color=color[2], label='fvm')
ax[4].plot(I*t/60./60., 1e-4*cDd[:,-1], color=color[3], label='fdm')
ax[4].plot(I*t/60./60., 1e-4*cB, color=color[1], label='analytical')



################################################################################
# plot integral flux at top interface into cathode over time
################################################################################

gradC = (cD[:,-2] - cD[:,-1])/(z[:,-2] - z[:,-1]) # first derivative
q = np.full(tPoints, 0.0)			# flux for each time step
qSum = np.full(tPoints, 0.0)			# integrated flux
oldQ = 0.0

for i in range(tPoints):
  newQ = oldQ + D*gradC[i]*A*deltaT*F
  qSum[i] = newQ
  q[i] = D*gradC[i]*A*deltaT*F
  oldQ = newQ

ax[5].plot([0,65/2.,65],[0,65/2.,65], 'o-', color=color[11], label='theory')
ax[5].plot(I*t/60./60., qSum/60./60., color=color[2], label='fvm')



################################################################################
# plot flux at top interface into cathode (for each time step)
################################################################################

ax[6].plot([0, I*tEnd/60./60./2., I*tEnd/60./60.], [I*deltaT/60./60., I*deltaT/60/60., I*deltaT/60./60.], 'o-', color=color[11], label='theory')
ax[6].plot(I*t/60./60., q/60./60., color=color[1], label='numerical')

dz = h0/nz

c1 = diffFiniteLayer(D, h0, j, N, n_el, t, h0)
c2 = diffFiniteLayer(D, h0, j, N, n_el, t, h0-dz)
gradC = (c2 - c1)/dz				# first derivative
ax[6].plot(I*t/60./60., -D*gradC*A*deltaT*F/60./60., label='analytical')



################################################################################
# plot amount of Li in the cathode
################################################################################

# we have nz points in z-direction, i.e. dz has size nz-1
# index is [time, location]

n = np.zeros(tPoints)		# mols of Li in whole cathode for each time step
for tt in range(tPoints):	# iterate over time
  for i in range(nz):		# iterate over dz
    n[tt] = n[tt] + dzAll[tt,i]*cD[tt,i+1]	# mols of Li add

n = n*A				# multiply with surface

ax[7].plot([0, C/60./60/2., C/60./60.], [0, C/F/2.0, C/F], 'o-', color=color[11], label='theory')
ax[7].plot(I*t/60./60., n, color=color[2], label='fvm')



################################################################################
# plot concentration profile
################################################################################

ax[8].plot(I*t/60./60., 1e3*dzAll[:,-1], label='fvm, upper')
ax[8].plot(I*t/60./60., 1e3*dzAll[:,-2], label='fvm, upper-1')
ax[8].plot(I*t/60./60., 1e3*dzAll[:,1], label='fvm, lower')



################################################################################
# plot concentration profile of last time step
################################################################################

#ax[9].plot(cD[-1,:],  1e3*z[3,:],  '-x', color=color[2], label='fvm')
#ax[9].plot(cDd[-1,:], 1e3*zd[3,:], '-x', color=color[3], label='fdm')

#ax[9].plot(I*t/60./60., 1e3*hIntermetallic(cToX(cD, T), dzAll))


################################################################################
# plot correction factor
################################################################################

# index is [time, location]
tcr = int(30.*60*60/I/deltaT)
#tcr = 2220

#for i in range(tcr, tcr+100,10):
#  ax[10].plot(1e4*(fAll[i,:]-1), 1e3*z[i,1:-1], 'o')


ax[11].plot(I*t/60./60., 1e4*(fAll[:,-1]-1), linewidth=0.5, label='fvm, top')
ax[11].plot(I*t/60./60., 1e4*(fAll[:,-2]-1), linewidth=0.5, label='fvm, top-1')



################################################################################
# save
################################################################################

ax[0].set_xlim(0,65)
ax[1].set_xlim(0,65)
ax[2].set_xlim(0,65)
#ax[3].set_xlim(0,1)
ax[4].set_xlim(0,65)
ax[5].set_xlim(0,65)
ax[6].set_xlim(0,65)
ax[7].set_xlim(0,65)
ax[11].set_xlim(0,65)


ax[0].set_ylim(0,1)
ax[1].set_ylim(0,6.8)
ax[2].set_ylim(0,0.75)
ax[4].set_ylim(0,1e-4*xToC(0.75, T))
ax[5].set_ylim(0,65)
ax[6].set_ylim(-0.1,0.1)
ax[7].set_ylim(0,3*nBi)
#ax[11].set_ylim(-1,1)


ax[0].set(xlabel=r"charge in Ah", ylabel=r"$E$ in mV")
ax[1].set(xlabel=r"charge in Ah", ylabel=r"$h$ in mm")
ax[2].set(xlabel=r"charge in Ah", ylabel=r"$x$")
ax[3].set(xlabel=r"$x$", ylabel=r"$z$ in mm")
ax[4].set(xlabel=r"charge in Ah", ylabel=r"$c$ in 10$^4$ mol/m$^3$")
ax[5].set(xlabel=r"charge in Ah", ylabel=r"influx charge in Ah (time-integrated)")
ax[6].set(xlabel=r"charge in Ah", ylabel=r"influx charge in Ah (each time step)")
ax[7].set(xlabel=r"charge in Ah", ylabel=r"n$_\mathrm{Li}$ in mol")
ax[8].set(xlabel=r"charge in Ah", ylabel=r"dz in mm")
ax[9].set(xlabel=r"charge in Ah", ylabel=r"$z$ in mm")
ax[10].set(xlabel=r"factor in 10$^{-4}$", ylabel=r"$z$ in mm")
ax[11].set(xlabel=r"charge in Ah", ylabel=r"factor in $10^{-4}$")


ax[0].legend(loc='lower left')
ax[1].legend()
ax[2].legend()
ax[3].legend()
ax[4].legend()
ax[5].legend()
ax[6].legend()
ax[7].legend()
ax[8].legend()
ax[9].legend()
ax[11].legend()


ax[0].set_title('cell potential')
ax[1].set_title('layer height over time')
ax[2].set_title('molare fraction at interface')
ax[3].set_title('profile molar fraction over height')
ax[4].set_title('concentration at interface over time')
ax[5].set_title('time-integrated flux into layer')
ax[6].set_title('flux into layer')
ax[7].set_title('amount of Li in layer')
ax[8].set_title('dz over time')
ax[9].set_title('thickness intermetallic phase')
ax[10].set_title('correction factor over height')
ax[11].set_title('correction factor over time')




f.subplots_adjust(left=0.12,bottom=0.11,right=0.98,top=0.98,wspace=0.15,hspace=0.25)
f.savefig("g.pdf")
