
# figure 3
import sys, numpy as np, scipy.constants, matplotlib.pyplot as plt
sys.path.append("../lib")
from emf import emf_x
from LiBi import *
from density import cToX
import density



################################################################################
# define materials and dimensions
################################################################################

# define dimensions
d1 = 0.090              # diameter bottom (Infos by Steffen)
d2 = 0.030              # upper diameter
A = np.pi*d1*d1/4.0     # surface area (bottom)
hE1 = 0.0025            # electrolyte layer height, start of discharge

# define operation parameters
I = 20.                 # current
j = I/A                 # current density
T = 420                 # °C
t0=60.9			# start of charge
t1=511.17		# start of discharge

# define initial condition
nBi = 1.248             # initial amount of Bi (Info Steffen)

# define discretisation
nz = 100                                        # cells in z
deltaT = 0.1                                    # time step
tEnd = 7.5*60
tPoints = int(tEnd/deltaT + 1)

# define material
D = 7.0e-9              # diffusion coefficient
n_el = 1                # number of electrons
F = scipy.constants.value('Faraday constant')

# read data
d = np.genfromtxt('0203_cycl_20A_-20A_7.5min_7.5min.txt', skip_header=1)



################################################################################
# get the initial values, start of charge
################################################################################

# get x from emf, calc density, volume and c0

print("\n*************** Initial values charge ***************\n")
print("T in °C: ", T, '\n')
x0 = 0.096
print("x0 (estimated): ", x0, '\n')

print("emf0, measured", d[0][1]/1000.)
print("emf0, calculated", emf_x(np.array([x0]), T)[0])
print(" ")

n0 = x0*nBi / (1.-x0)
print("n0 (Li): ", n0)
print("n0 (Bi): ", nBi, '\n')

rho = density.rhoLiBi_x(x0, T)
print("rho:  ", rho, '\n')

# volume is needed for concentration

V0 = (n0*MLi() + nBi*MBi())/rho
h0 = V0 / A
print("V: ",  V0)
print("h: ", h0, '\n')

# get concentrations
cBi = nBi / V0
print("c0 (Bi)", cBi)
c0 = n0 / V0
print("c0 (Li)", c0)

# compute x0 by formula
c0calc = rho * x0 / (x0*MLi() + (1.-x0)*MBi())
print("c0, calculated from x0: ", c0calc, '\n')
print(" ")



################################################################################
# get the initial values, start of discharge
################################################################################

print("\n*************** Initial values discharge ***************\n")

# get concentration change due to charge
Idt = I * tEnd       # 7.5 min
n1 = Idt / (F*n_el)     # amount of Li entering the cathode during discharge
x1 = (n0-n1)/(n0-n1+nBi)# molar fraction Li
c1 = density.rhoLiBi_x(x1, T) * x1/(x1*MLi() + (1.-x1)*MBi()) # concentration Li

print("x (computed): ", x1)
print("c (computed): ", c1)
print(" ")
print("emf(x1, computed): ", emf_x(np.array([x1]), T)[0])
print(" ")

V1 = (n1*MLi() + nBi*MBi())/density.rhoLiBi_x(x1, T)    # get volume for layer thickness
h1 = V1 / A
print("V: ",  V1)
print("h: ", h1, '\n')
hE0 = hE1 - (h0-h1)                                     # height electrolyte at very beginning
print("h(electrolyte): ", hE0)



################################################################################
# plot measurement
################################################################################

helmholtzBlau          = (       0.,  88./255., 156./255.)
rossendorfOrange       = (207./255., 104./255., 0.)
helmholtzDunkelBlau    = (       0.,  62./255., 110./255.)
grau1                  = ( 81./255.,  81./255.,  81./255. )
grau2                  = (156./255., 156./255., 156./255.)
grau3                  = (185./255., 185./255., 185./255.)

f, ax = plt.subplots()
f.set_size_inches(4,4)

ax.xaxis.label.set_size(12)
ax.yaxis.label.set_size(12)
ax.tick_params(axis='x', labelsize=10)
ax.tick_params(axis='y', labelsize=10)


plt.plot(d[:,0]/60., 1e-3*d[:,1], color=grau3, linewidth=5, label='experiment')



################################################################################
# get ohmic overpotential
################################################################################

from LiBi import sigmaLiClFI
from ohm import etaCylinder
sigma = sigmaLiClFI(T)
print("Conductivity: ", sigma)
eta0 = etaCylinder(I, sigma, d1, d2, hE0)
eta1 = etaCylinder(I, sigma, d1, d2, hE1)
print("Conductivity: ", sigma)
print("Ohmic loss (start charge): ", eta0)
print("Ohmic loss (start discharge): ", eta1)



################################################################################
# plot charge curve
################################################################################

# start up -> start charge
t = np.array([0, t0, t0])
e0 = emf_x(cToX(c0, T), T)
e=np.array([e0, e0, e0+eta0])
plt.plot(t/60., e, linewidth=1, color='k')

# charge
t = np.linspace(0, tEnd, 100)
Idt = I * t             # charge
n = n0 - Idt / (F*n_el) # amount of Li
x = n / (n+nBi)         # molar fraction Li
c = density.rhoLiBi_x(x, T) * x/(x*MLi() + (1.-x)*MBi()) # concentration Li

V = (n*MLi() + nBi*MBi())/density.rhoLiBi_x(x, T)       # get volume for layer thickness
h = V / A                                               # thickness cathode
hE = hE1 - (h - h1)                                     # Li is removed; cathode gets thinner, electrolyte thicker

e = emf_x(cToX(c, T), T) + etaCylinder(I, sigma, d1, d2, hE)
plt.plot((t0+t)/60., e, linewidth=1, color='k')

# end of charge
t = np.array([t1, t1])
e1 = emf_x(cToX(c1, T), T)
e = np.array([e1+eta1, e1])
plt.plot(t/60., e, linewidth=1, color='k')

# start of discharge
t = np.array([t1, t1])
e = np.array([e1, e1-eta1])
plt.plot(t/60., e, linewidth=1, color='k')



################################################################################
# plot analytical solutions
################################################################################

# plot analytical solution for infinite layer
#from diffAnalytical import diffSeminInfinite
#t = np.linspace(0.0, 7.5*60, 100)                # time
#cA = c1 + c_A_SemiInf(D, h1, j, n_el, t, h1)
#emf = emf_x(cToX(cA, T), T) - hE1*eta
#plt.plot((t+t1)/60.,1e3*emf,'+',linewidth=1,label='infinite layer')

# plot analytical solution for finite layer
from diffAnalytical import diffFiniteLayer
N = 1000 # number of summations for integral
t = np.linspace(0.0, 7.5*60, 100)                # time
cB = c1 + diffFiniteLayer(D, h1, j, N, n_el, t, h1)
emfB = emf_x(cToX(cB, T), T) - eta1
#plt.plot((t+t1)/60., 1e3*emfB, '-', linewidth=1, label='analytical solution')


################################################################################
# plot numerical solutions
################################################################################

# plot numerical solution
from fvm import c_numerical
t = np.linspace(0.0, tEnd, tPoints)
#cC,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T)
#emfC = emf_x(cToX(cC[:,-1],T),T) - hE1*eta
#plt.plot((t+t1)/60.,1e3*emfC,linewidth=0.5,label='numerical')

# with volume change
t = np.linspace(0.0, tEnd, tPoints)
#cD,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True)
#deltaH = z[:,-1] - h1   # height change for each time step
#emfD = emf_x(cToX(cD[:,-1],T),T) - (hE1 - deltaH)*eta
#plt.plot((t+t1)/60.,1e3*emfD,'--',linewidth=1,label='vol. change')

# with variable diffusivity
t = np.linspace(0.0, tEnd, tPoints)
#cE,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, varD=True)
#emfE = emf_x(cToX(cE[:,-1],T),T) - hE1*eta
#plt.plot((t+t1)/60.,1e3*emfE,linewidth=0.5,label='variable D')

# with vol change and variable D
t = np.linspace(0.0, tEnd, tPoints)
cF,z = c_numerical(D, h1, j, MLi(), n_el, tEnd, deltaT, nz, c1, T, volChange=True, varD=True)[:2]
deltaH = z[:,-1] - h1     # height change for each time step
emfF = emf_x(cToX(cF[:,-1], T), T) - etaCylinder(I, sigma, d1, d2, hE1-deltaH)
plt.plot((t+t1)/60., emfF, linewidth=1, label='numerical model', color='k')



################################################################################
# save
################################################################################

#plt.ylim(550,700)
plt.xlim(0,20)

ax.set(xlabel=r"$t$ in min", ylabel=r"$E$ in V")
plt.legend()

f.subplots_adjust(left=0.14,bottom=0.11,right=0.98,top=0.98,wspace=0.15,hspace=0.25)
f.savefig("f1.pdf")
