""" Provides different density functions """

import numpy as np
from LiBi import MLi, MBi

def linear_x(x, rho1, rho2):
    """ Compute density as linear function of the molar fraction """
    return (rho1-rho2)*x + rho2

def idealSolution_x(x,M1,M2,rho1,rho2):
    """ Compute density for an ideal solution using molar fraction """
    return (x*M1 + (1.-x)*M2) / (x*M1/rho1 + (1.-x)*M2/rho2)

def Vegard_x(x,M1,M2,rho1,rho2):
    """ Compute density using Vegard's law and molar fraction """
    return (x*M1 + (1.-x)*M2) / (x*(M1/rho1)**(1./3.) + (1.-x)*(M2/rho2)**(1./3.))**3

def rhoLi(T):
    """ Compute density of Lithium with T in °C"""
    # Smithells metals reference book, p. 14-10
    T = T + 273.15
    rhoM = 0.518e3
    dRhoDt = -0.1
    Tm = 453.5
    return rhoM + dRhoDt*(T - Tm)

def rhoBi(T):
    """ Compute density of Bismuth with T in °C """
    # Smithells metals reference book, p. 14-10
    T = T + 273.15
    rhoM = 10.05e3
    dRhoDt = -1.18
    Tm = 544.
    return rhoM + dRhoDt*(T - Tm)

def rhoLiBi_x(x,T):
    """ Compute density of Li-Bi alloy, fitted by molar fraction
    - T in °C
    """
    a1 = 0.93836162 
    a2 = 1.64296296
    return (x*MLi() + (1.-x)*MBi()) / (x**a1 * MLi()/rhoLi(T) + (1.-x)**a2 *MBi()/rhoBi(T))

# density alloy, fitted by molar concentration
def rhoLiBi_c(c,T):
    """ Compute density of Li-Bi alloy, fitted by molar concentration
    - T in °C
     """

    # function should accept floats and numpy array
    c = np.asarray(c)
    # set negative concentrations to zero
    #c[c < 0] = 0

    a = 1.45010274e+00
    b = 6.69579965e-01 
    d = 8.13105782e-09 
    e = 3.70287809e+01
    f = 2.37129019e+00 
    g = 6.95494289e-01

    cInf = rhoLi(T) / MLi()
    # limit c to cInf
    #c[c > cInf] = cInf
    return (c**f * d + (cInf - c)**g * e) / ( (c/cInf)**a * cInf**f * d/rhoLi(T) + ((cInf-c)/cInf)**b * cInf**g * e/rhoBi(T))

def xToC(x,T):
    """ Convert molar fraction to concentration, with T in °C """
    return rhoLiBi_x(x, T)*x / (x*MLi() + (1.-x)*MBi())

def cToX(c,T):
    """ Convert molar concentration to fraction, with T in °C """
    return c*MBi() / (rhoLiBi_c(c, T) + c*MBi() - c*MLi())
