""" Solution of the diffusion equation by analytical models """

import scipy.special
import numpy as np
import scipy.constants

F = scipy.constants.value("Faraday constant")

# error function, needed for analytical solution of diffusion in 
# semi-infinite layer
def ierfc(x):
    """ Compute error function, which is needed
    for the solution in a semi-infinite layer
    """
    return (np.exp(-(x**2))/np.pi**0.5) - (x*scipy.special.erfc(x))

def fSum(D, h, n, t, z):
    """ Compute sum term in the solution of the 
    diffusion equation for a finite layer
    """
    a1 = (-1)**n / n**2
    a2 = np.exp(-(D*n**2*np.pi**2*t)/h**2)
    a3 = np.cos(n*np.pi*z/h)
    return a1*a2*a3

def diffFiniteLayer(D, h, j, N, n_el, t, z):
    """ Compute solution of diffusion equation for a finite layer
    - D: diffusion coefficient
    - h: height of layer
    - j: current density
    - N: number of summands to use (the higher the more accurate)
    - n_el: number of exchanged electrons
    - t: time
    - z: coordinate
     """
    # first summand
    a1 = D*t/h**2
    # second summand
    a2 = (3.*z**2 - h**2)/(6.*h**2)
    # last sum
    a3 = 0
    for n in range(1, N):
        a3 = a3 + fSum(D, h, n, t, z)

    return (j*h/(n_el*F*D))*(a1 + a2 - (2./np.pi**2)*a3)


def diffSemiInfinite(D, h, j, n_el, t, z):
    """ Compute solution of diffusion equation for semi-infinite layer
    - D: diffusion coefficient
    - h: height of layer
    - j: current density
    - n_el: number of exchanged electrons
    - t: time
    - z: coordinate
    """
    return (2.*j*(D*t)**0.5)/(n_el*F*D) * ierfc(np.divide(h - z, (2.*(D*t)**0.5), where=t!=0))
