""" Computes electromotive force for the Li-Bi couple """

import numpy as np
import phaseDiagramme as pd


def emf_liq_x(x, T):
    """ Compute electromotive force in the liquid phase only
    - T: temperature in °C
    - x: molar fraction
    """
    T=T+273.15 # convert to K

    # function should accept floats and numpy array
    x = np.asarray(x)
    # do not allow x to be zero or negative
    x[x <= 1e-12] = 1e-12
    
    o=0.7866572396802591
    p=-0.0060998993260001815

    a=-7.307133461698026e-05
    b=0.0046614686350452544
    c=-0.01650070632108194
    d=0.02896169966609987
    e=-0.02300638772905505
    f=0.001746008388289424
    
    return o + p*np.log(x) + T * (a*np.log(x) + b*x + c*x*x + d*x*x*x + e*x*x*x*x + f*x*np.log(x))

def emf_x(x, T):
    """ Compute open-circuit voltage depending on molar fraction and temperature in °C """

    # function should accept floats and numpy array
    x = np.asarray(x)
    emf = np.zeros(x.shape)
    xMin = 0.01

    # get phase diagramme; xlig1 is left branch, 2 right branch
    Tliq, xliq1, xliq2 = pd.Tliq()

    ##### liquid area #####

    # liquid, xMin<x<0.75 (region L)
    emf = np.where(np.logical_and(np.logical_and(T > Tliq(x), x<0.75), x>=xMin), emf_liq_x(x,T), 0.0)	# all others are zero

    # liquid, x < xMin: use value of x=0.01 (region L)
    emf = np.where(np.logical_and(T > Tliq(x), x<xMin), emf_liq_x(xMin,T), emf)


    ##### two-phase area #####

    # two-phase area left: below liquidus and x<0.14 (region I ... and solid below)
    # obviously, numpy first computes both functions for all values, and decides after that which value to take
    emf = np.where(np.logical_and(T <= Tliq(x), x<0.14), emf_liq_x(xliq1(-T),T), emf)

    # two-phase area of Li3Bi, right below liquidus: 0.14 < x < 0.73,  above 688K (region IV ... and solid below)
    emf = np.where(np.logical_and(np.logical_and(T <= Tliq(x), x>=0.14), x<=0.73), emf_liq_x(xliq2(T),T), emf)

    # two-phase area of LiBi intermetallic right below liquidus: x <= 0.5 , below 688K (region II and III)
    emf = np.where(np.logical_and(np.logical_and(np.logical_and(T<Tliq(x), x>=0.14), np.logical_and(x<=0.5, T>=243)), T<=415), emf_liq_x(xliq2(T), T) - (-0.000359576*(T+273.15) + 0.246034967), emf)

    # two-phase, between 0.73 and 0.75 (region V ... and all below)
    emf = np.where(np.logical_and(x<=0.75, x>=0.73), emf_liq_x(xliq2(T),T) + ((x - 0.73) / (0.75 - 0.73)) * (0 - emf_liq_x(xliq2(T), T)), emf)

    ##### solid area #####

    # solid two-phase region of LiBi-Li3Bi intermetallic: 0.5 < x < 0.73, below 688K (region SIV and SII)
    # -> No measured voltage known, modeled as 0.0 V
    emf = np.where(np.logical_and(np.logical_and(x>=0.5, x<=0.75), T < 415.), 0.0, emf)

    # solid, below 516K / 243°C (region SI, SII ... and right of these)
    emf = np.where(T <= 243, 0, emf)

    # x above 0.75 (region VI, VII and SIII)
    emf = np.where(x >= 0.75, 0, emf)

    return emf
