""" Finite difference code for solving the diffusion equation
- 2nd order in space
- 1st order in time

- central differencing is used
- concentration (c) and diffusivity (D) are defined at points
- distance (dz) is defined between points
- to use central differencing, an additional extrapolated point for (D) is added on both sides
- similarly dz is added on both sides
- these extrapolated D and dz-values are used only for compact writing
- they are overwritten by the boundary condition
"""

import numpy as np
from density import rhoLiBi_c
from LiBi import D_c
import scipy.constants 
F = scipy.constants.value("Faraday constant")

def calcA(D, dt, dz):
    """ Compute the matrix A of A*x = b """

    # matrix definition
    dzu = dz[1:]                        # dz starts before point, so upper is the second element
    dzl = dz[:-1]                       # and lower goes to last - one
    Di  = D[1:-1]                       # where c is defined (inner)
    Du  = D[2:]                         # at i+1 (upper)
    Dl  = D[:-2]                        # at i-1 (lower)

    a = dt / ((dzu + dzl)*dzu*dzl)         		# factor for matrix coefficient

    udiagonal = -(Di+Du)*dzl*a                          # upper subdiagonal
    diagonal  = 1. + ((Di+Du)*dzl + (Dl+Di)*dzu)*a      # diagonal elements
    ldiagonal = -(Dl+Di)*dzu*a                          # lower subdiagonal

    # thinking LINEwise, the upper diagonal never reaches the last line
    udiagonal = udiagonal[:-1]                          # rm last element
    ldiagonal = ldiagonal[1:]                           # rm first element

    AA = (                                              # set up matrix
        np.diagflat(diagonal)
        + np.diagflat(ldiagonal, -1)
        + np.diagflat(udiagonal, 1)
    )

    # boundary condition: no flux on the bottom
    AA[0, 0] = -1
    AA[0, 1] = 1

    # boundary condition: Faraday flux on top (second order)
    a = dz[-2]*(dz[-2] + dz[-3])*dz[-3]
    AA[-1, -3] = -(dz[-2])**2 / a
    AA[-1, -2] = (dz[-2]+dz[-3])**2 / a
    AA[-1, -1] = -((dz[-2]+dz[-3])**2 - (dz[-2])**2) / a

    # boundary condition: Faraday flux on top (first order)
    #AA[-1, -2] = 1./dz[-2]
    #AA[-1, -1] = -1./dz[-2]

    return AA

# solves the diffusion equation numerically for molar concentration
def c_numerical(D0, h, j, M_A, n_el, tEnd, dt, Np, c0, T, volChange=False, varD=False):
    """ Solve the diffusion equation for molar concentration
    - D0: diffusion coefficient
    - h: layer height
    - j: current density
    - M_A: molar mass of diffusing species
    - n_el: number of exchanged electrons
    - tEnd: end time
    - dt: time step
    - Np: number of points
    - c0: initial concentration
    - T: temperature in °C
    - volChange: volume change
    - varD: variable diffusivity
    """

    # time discretisation
    Nt = int((tEnd / dt) + 1)  		# number of discretisation points in time

    # space discretization
    dz = np.full((Np+1), h / (Np - 1))	# distance between points; incl. two outside-distances
    dz[0] = 1e6                         # the extrapolated values will be overwritten by the boundary condition
    dz[-1] = 1e6                        # 1e6 is only a safety measure
    dz0 = dz.copy() 			# save initial space discretisation
    dzOld = dz.copy()			# distance of last time step
    dzAll = dz.copy()			# dz of all time steps

    # initial condition
    c = c0 * np.ones(Np)		# concentration of current time step
    cOld = c.copy()			# concentration of last time step
    cAll = c.copy()			# concentration of all time steps

    # vales, defined at dz, i.e. BETWEEN points
    rho0 = np.zeros(Np-1)		# initial density, only defined between points
    for i in range(rho0.size):		
        cMean = (c[i] + c[i+1]) / 2.0	# dz 0 is between point 0 and 1
        rho0[i] = rhoLiBi_c(cMean, T)
    m = rho0*dz0[1:-1]			# "mass" with unit mass/surface
    mOld = m.copy()

    V0 = m/rho0               		# initial "volume" (in reality a height in m)
    Vold = V0.copy()			# volume of old time step

    # values, defined AT points
    rho0p = rhoLiBi_c(c, T)             # initial density at points
    dz0p = (dz[1:-2]+dz[2:-1])/2.0	# "size" of point
    dz0p = np.append(dz[1]/2.0, dz0p)   # first point owns only half a volume
    dz0p = np.append(dz0p, dz[-2]/2.0)	# last point has only half volume
    
    mp = rho0p*dz0p                     # "mass" with unit mass/surface
    mOldp = mp.copy()

    V0p = mp/rho0p			# initial "volume" (in reality a height in m)
    Voldp = V0p.copy()                  # volume of old time step

    # diffusion coefficient
    D = np.full(Np+2, D0)		# includes two outside points: needed for interpolation
    if(varD):
        D[1:-1] = D_c(c)		# update by concentration
        D[0]  = 1e6			# extrapolated values will be overwritten by boundary condition
        D[-1] = 1e6			# 1e6 only serves to see the mistake, if they are not overwritten

    # matrix definition
    AA = calcA(D, dt, dz)

    bb0 = np.zeros_like(c)           	# Term b in A*c=b
    dndt = j / (n_el*F)                 # molar flux
    bb0[0] = 0                          # zero mass flux at bottom boundary
    bb0[-1] = -dndt / D[-2]
    # dz is inner distance; D is last point - 1, because it has more points than c

    # time cycle
    for kk in range(1, Nt):
	# compute matrices
        AA = calcA(D, dt, dz)		# obtain term A of matrix A*c=b
        bb = cOld 			# term b is the concentration of last time step
        bb[0] = 0			# no flux on bottom
        bb[-1] = -dndt / D[-2]		# Faraday flux at top (D at point 2 to skip extrapolated point)

        c = np.linalg.solve(AA, bb)	# solve and save concentration

        if(varD):			# update diffusion coefficient for next iteration
            D[1:-1] = D_c(c)
            D[0]  = 1e6			# extrapolated 'D's are not used / overwritten by boundary condition
            D[-1] = 1e6			# large value to see if we make a mistake


        if(volChange): 			# change dz only, if volume change is switched on

            # FIRST: scale volume/dz -> scaling factor is compute BETWEEN points

            # interpolate concentration BETWEEN points
            cl = np.append(c[0],c)	# lower/left; first value is extrapolated
            cu = np.append(c,c[-1])	# upper/right; last value is extrapolated
            cBetween = (cu + cl) / 2.0	# concentration between points

            # get distance between 'cBetween'
            dzu = dz[1:]                # dz starts before point, so upper is the second element
            dzl = dz[:-1]               # and lower goes to last - one
            dzBetween = 0.5*(dzu+dzl)	# distance 'over' the points

            # compute mass flux - defined AT points
            # mass flux is positive, when in positive z-direction
            mflux = -D[1:-1] * M_A * (cBetween[1:] - cBetween[:-1]) / dzBetween
            mflux[0] = 0 			# mass flux at bottom
            mflux[-1] = -j*M_A / (n_el*F)  	# mass flux at electrolyte; - because flowing DOWNwards

            # compute mass, density, volume BETWEEN points
            # accumulation of mass means: + at lower end, - at upper interface
            m1 = mOld + (mflux[:-1] - mflux[1:]) * dt
            rho = rhoLiBi_c(cBetween[1:-1], T)
            V  = m1/rho
            f1 = V/Vold


            # SECOND: scale concentration -> scaling factor is defined AT points

            Df = (D[1:-2] + D[2:-1])/2.0 			# D between points (rm 2 outer extrapolated points)
            mflux = -Df * M_A * (c[1:] - c[:-1])/dz[1:-1] 	# mass flux between points
            mflux = np.append(0, mflux)				# mass flux below points
            mflux = np.append(mflux, -j*M_A / (n_el*F))		# mass flux on top of geometry
            mp = mOldp + (mflux[:-1] - mflux[1:]) * dt 		# mass at points
            rhop = rhoLiBi_c(c, T)				# density at point
            Vp = mp/rhop					# volume at point
            fp = Vp/Voldp					# scaling factor at point

            # scale dz
            dz[1:-1] = dz0[1:-1] * (V/V0)

            # scale concentration
            c = c / fp
            c[0] = c[1]

            Vold = V
            Voldp = Vp
            dz[0] = 1e6
            dz[-1] = 1e6
            dzOld = dz			# save dz for next time step

            mOld = m1
            mOldp = mp

        cOld = c			# save c for next time step
        cAll = np.vstack((cAll,c))	# save concentration profile
        dzAll = np.vstack((dzAll,dz))	# save dz

    # compute z-coordinate for each time step
    z = np.zeros((dzAll.shape[0], Np))	# height of each point for all times
    z0 = np.zeros(dzAll.shape[0])	# only temporary variable
    i = 1				# start with second dz (to skip extrapolated dz)

    while i < len(dz)-1:		# do not count first and last dz
        z0 = z0+dzAll[:,i]		# add height of new points
        z[:,i] = z0			# save
        i = i + 1
    return cAll, z
