""" Finite volume code to solver diffusion equation
- 2nd order in space
- 1st order in time - Euler

- concentration (c), diffusivity (D) and size of the cell (dz) are defined at cell centre
- distance of cell-centres (delta) and face-centred diffusivity (Df) defined on faces
"""

import numpy as np
from density import rhoLiBi_c
from LiBi import D_c
import scipy.constants 
F = scipy.constants.value("Faraday constant")

def delta(dz):
    """ Compute distance between points """
    d = (dz[:-1] + dz[1:])/2.0      	# distance between points (delta)
    d = np.append(dz[0]/2.0, d) 	# first point: distance to face
    d = np.append(d, dz[-1]/2.0)	# last point: distance to face
    return d

def Df(D):
    """ Interpolate diffusivity from cell centre to face """
    Df = (D[:-1] + D[1:]) / 2.0         # diffusion coefficient on face
    Df = np.append(D[0], Df)            # add extrapolated at beginning
    Df = np.append(Df, D[-1])           # and end
    return Df

def calcA(D, dt, dz):
    """ Compute matrix A of A*c = b """

    a = Df(D)*dt/delta(dz)		# factor for matrix coefficient - defined for each face

    udiagonal = a[1:]                   # upper subdiagonal (first face is between cell 1+2; index 1)
    diagonal  = -dz -a[:-1] - a[1:]     # diagonal elements
    ldiagonal = a[:-1]                  # lower subdiagonal

    # thinking LINEwise, the upper diagonal never reaches the last line
    udiagonal = udiagonal[:-1]          # rm last element
    ldiagonal = ldiagonal[1:]           # rm first element

    AA = (                              # set up matrix
        np.diagflat(diagonal)
        + np.diagflat(ldiagonal, -1)
        + np.diagflat(udiagonal, 1)
    )

    # boundary condition: no flux on the bottom
    AA[0, 0] = AA[0, 0] + a[0]

    # Faraday flux on top
    AA[-1, -1] = AA[-1, -1] + a[-1]

    return AA

def c_numerical(D0, h, j, M_A, n_el, tEnd, dt, Np, c0, T, volChange=False, varD=False):
    """ Solve the diffusion equation for molar concentration
    - D0: diffusion coefficient
    - h: layer height
    - j: current density
    - M_A: molar mass of diffusing species
    - n_el: number of exchanged electrons
    - tEnd: end time
    - dt: time step
    - Np: number of points
    - c0: initial concentration
    - T: temperature in °C
    - volChange: volume change
    - varD: variable diffusivity
    """

    # time discretisation
    Nt = int((tEnd / dt) + 1)  		# number of discretisation points in time

    # space discretization
    dz = np.full(Np, h/Np)		# size of control volumes
    dz0 = dz.copy() 			# save initial space discretisation
    dzOld = dz.copy()			# distance of last time step
    dzAll = dz.copy()			# dz of all time steps

    # initial condition
    c = c0 * np.ones(Np)		# concentration of current time step
    cOld = c.copy()			# concentration of last time step
    cAll = c.copy()			# concentration of all time steps

    fAll = np.ones(Np)
    f = fAll.copy()

    # density and volume
    rho0 = rhoLiBi_c(c, T)		# density at the beginning (kg/m3)
    m0 = rho0*dz			# "mass" with unit mass/surface (inital value)
    m = m0.copy()			# "mass" with unit mass/surface (of current time step)
    mOld = m0.copy()			# mass of old time step

    V0 = m0/rho0               		# initial "volume" (in reality a height in m)
    V = V0
    Vold = V0.copy()			# volume of old time step
    mflux = np.zeros(Np)	 	# mass flux at points in kg/m2/s

    D = np.full(Np, D0)			# diffusion coefficient
    if(varD):
        D = D_c(c)			# update by concentration

    # matrix definition
    AA = calcA(D, dt, dz)

    bb0 = c.copy()*dz	        	# Term b in A*c=b
    dndt = j / (n_el*F)                 # molar flux
    bb0[-1] = bb0[-1] -dndt * dt	# Faraday flux boundary condition

    # counter can be used to change volume every nth time step, only
    #count = 0
    # time cycle
    for kk in range(1, Nt):

        #count = count + 1
        #if(count > 1):			# 27 is the critical value
        #  count = 0

	# compute matrices
        AA = calcA(D, dt, dz)		# obtain term A of matrix A*c=b
        bb = -cOld*dz 			# term b is the concentration of last time step * cell volume
        bb[-1] = bb[-1] - dndt*dt	# Faraday flux at top

        #c = np.linalg.solve(AA, bb)	# solve and save concentration

        AA = np.nan_to_num(AA)		# rm nan (necessary for least squares solver)
        bb = np.nan_to_num(bb)
        c = np.linalg.lstsq(AA, bb)[0]	# solve using least squares

        if(varD):			# update diffusion coefficient for next iteration
            D = D_c(c)

        if(volChange): 			# change dz only, if volume change is switched on

            # compute mass flux - defined at faces
            # mass flux is positive, when in positive z-direction
            gradC = (c[1:] - c[:-1])/delta(dz)[1:-1]
            gradC = np.append(0.0, gradC)		# zero gradient on bottom
            gradC = np.append(gradC, j/(n_el*F*D[-1]))	# pos. because upper c is large than lower; or because positive charges (Li+) go downwards i.e. inverse to coordinate system -> this cancels the minus out

            mflux = -Df(D) * M_A * gradC 		# kg/m2/s (was not multiplied by surface)

            # compute mass, density, volume BETWEEN points
            # accumulation of mass means: + at lower end, - at upper interface
            m = mOld + (mflux[:-1] - mflux[1:])*dt
            rho = rhoLiBi_c(c, T)
            V = m/rho
            f = V/Vold

            # scale dz and c
            if(True):#(count == 1):
                dz = dzOld*f
                #dz = dz0*(V/V0)
                c = c/f
                Vold = V
                dzOld = dz
            mOld = m

        cOld = c			# save c for next time step
        cAll = np.vstack((cAll,c))	# save concentration profile
        dzAll = np.vstack((dzAll,dz))	# save dz
        fAll = np.vstack((fAll,f))

    # compute z-coordinate for each time step
    z = np.zeros((dzAll.shape[0], Np+2))# height of each point for all times (2 boundary points added)
    z0 = np.zeros(dzAll.shape[0])	# only temporary variable
    z0 = dzAll[:,0]/2.0			# height of second point
    z[:,1] = z0				# (first is at boundary with z=0)
    
    i = 1				# start with second dz (first is distance boundary - point 1)

    while i < len(dz-1):			# do not count first and last dz (distances to boundary)
        z0 = z0+(dzAll[:,i]+dzAll[:,i-1])/2.0	# add height of new points
        z[:,i+1] = z0				# save
        i = i + 1

    z[:,-1] = z[:,-2] + dzAll[:,-1]/2.0		# add height of upper boundary point

    # add boundary concentration
    # at point concentration of point 0 as boundary value (zero gradient)
    cAll = np.hstack((np.reshape(cAll[:,0], (-1,1)), cAll))
    cAll = np.hstack((cAll, np.reshape(cAll[:,-1] + j*(dzAll[:,-1]/2.0)/(n_el*F*D[-1]), (-1,1))))
    # at time 0, concentration is c0 everywhere (overwrite influx BC)
    cAll[0,-1] = c0
    return cAll, z, fAll, dzAll
